"""
    Ramachandran extension

    Print both the phi and psi angles to standard out.  See the individual
    functions for more info.

    Author:  Mike Bradley and Todd Dolinsky
"""

__date__ = "17 February 2006"
__author__ = "Mike Bradley, Todd Dolinsky"

from src.utilities import getDihedral
import extensions


def addExtensionOptions(extensionGroup):
    """
    Add options to set output type.
    """
    extensionGroup.parser.set_defaults(rama_output="rama")
    extensionGroup.add_option(
        "--phi_only",
        dest="rama_output",
        action="store_const",
        const="phi",
        help="Only include phi angles in output. "
        + "Rename output file {output-path}.phi",
    )

    extensionGroup.add_option(
        "--psi_only",
        dest="rama_output",
        action="store_const",
        const="psi",
        help="Only include psi angles in output. "
        + "Rename output file {output-path}.psi",
    )


def usage():
    return "Print the per-residue phi and psi angles to {output-path}.rama for Ramachandran plots"


def create_rama_output(routines, outfile, outputtype="rama"):

    routines.write(
        "\nPrinting %s angles for each residue...\n"
        % (outputtype if outputtype != "rama" else "phi and psi")
    )
    verboseHeader = "Residue        %s\n" % (
        outputtype.capitalize() if outputtype != "rama" else "Phi          Psi"
    )
    routines.write(verboseHeader)
    routines.write("-" * len(verboseHeader) + "\n")

    output = extensions.extOutputHelper(routines, outfile)

    # Initialize some variables

    protein = routines.protein

    for residue in protein.getResidues():
        if residue.hasAtom("N"):
            ncoords = residue.getAtom("N").getCoords()
        else:
            continue

        if residue.hasAtom("CA"):
            cacoords = residue.getAtom("CA").getCoords()
        else:
            continue

        if residue.hasAtom("C"):
            ccoords = residue.getAtom("C").getCoords()
        else:
            continue

        try:
            if residue.peptideN != None:
                pepncoords = residue.peptideN.getCoords()
            else:
                continue

            if residue.peptideC != None:
                pepccoords = residue.peptideC.getCoords()
            else:
                continue
        except AttributeError:  # Non amino acids
            continue

        output.write(str(residue))

        if outputtype in ("rama", "phi"):
            phi = getDihedral(pepccoords, ncoords, cacoords, ccoords)
            output.write("\t%.4f" % phi)

        if outputtype in ("rama", "psi"):
            psi = getDihedral(ncoords, cacoords, ccoords, pepncoords)
            output.write("\t%.4f" % psi)

        output.write("\n")

    routines.write("\n")


def run_extension(routines, outroot, options):
    """
    Print the list of phi and psi angles for use in a Ramachandran plot.

    Parameters
        routines:  A link to the routines object
        outroot:   The root of the output name
        options:   options object
    """
    outputType = options.rama_output
    outname = outroot + "." + outputType
    with open(outname, "w") as outfile:
        create_rama_output(routines, outfile, outputtype=outputType)



Import("env")
Import("compile_targets")

from os.path import dirname, basename, join, isfile
from os import walk
from glob import iglob

compile_target_files = [str(x[0]) for x in compile_targets]


def getAllFiles(root_dir, python_only=False, recursive=True):
    if recursive:
        for root, dirnames, filenames in walk(root_dir):
            for filename in filenames:
                if not python_only or filename.endswith(".py"):
                    yield join(root, filename)

    else:
        for filename in iglob(join(root_dir, "*")):
            if not isfile(filename):
                continue

            if not python_only or filename.endswith(".py"):
                yield filename


def installFile(file_name, build_target="install"):
    # Don't do compile targets
    if file_name in compile_target_files:
        return
    # ignore pyc and pyo. We get them from the py build.
    if file_name.endswith(".pyc") or file_name.endswith(".pyo"):
        return

    if basename(file_name).startswith("."):
        return

    target = File(file_name)
    if file_name.endswith(".py"):
        compiled_name = file_name + "c"
        result = env.Command(
            compiled_name, file_name, CompilePython("$TARGET", "$SOURCE")
        )
        if GetOption("clean"):
            env.Default(result)
        Alias(build_target, env.Install(env["PREFIX"] + dirname(file_name), result))
    else:
        Alias(build_target, env.Install(env["PREFIX"] + dirname(file_name), target))


# Contrib
# for file_name in getAllFiles('contrib/ZSI-2.1-a1/ZSI'):
# 	installFile(file_name)
#
# installFile('contrib/ZSI-2.1-a1/Copyright')

# ProPKA
for file_name in getAllFiles("propka30/"):
    installFile(file_name)

# Whole directories
for dir_name in ("dat/", "doc/", "examples/", "jmol/", "images/", "ZSI/", "3dmol/"):
    for file_name in getAllFiles(dir_name):
        installFile(file_name)

# Compiled Targets
for target in compile_targets:
    file_name = str(target[0])
    if file_name.endswith(".py"):
        compiled_name = file_name + "c"
        result = env.Command(
            compiled_name, file_name, CompilePython("$TARGET", "$SOURCE")
        )
        if GetOption("clean"):
            env.Default(result)
        Alias("install", env.Install(env["PREFIX"] + dirname(file_name), result))
    else:
        Alias("install", env.Install(env["PREFIX"] + dirname(file_name), target))

# PDB2PKA
for file_name in getAllFiles("pdb2pka/"):
    installFile(file_name)

# Main Program
for dir_name in ("src/", "extensions/"):
    for file_name in getAllFiles(dir_name, python_only=True):
        installFile(file_name)

Alias("install", env.Install(env["PREFIX"], "pdb2pqr.py"))
Alias("install", env.Install(env["PREFIX"], "pdb2pqr.cgi"))
Alias("install", env.InstallAs(env["PREFIX"] + "/index.html", "html/server.html"))

for file_name in (
    "AppService_client.py",
    "AppService_services.py",
    "AppService_services_types.py",
    "AppService_types.py",
    "AUTHORS",
    "ChangeLog.md",
    "COPYING",
    "main.py",
    "main_cgi.py",
    "NEWS",
    "pka.py",
    "pdb2pqr.css",
    "README.md",
):
    installFile(file_name)

Alias(
    "install",
    env.Command(
        env["PREFIX"] + "tmp", None, [Mkdir("$TARGET"), Chmod("$TARGET", 0777)]
    ),
)

# Testing stuff to test installed pdb2pqr
for dir_name in ("tests/", "scons/", "site_scons/"):
    for file_name in getAllFiles(dir_name):
        installFile(file_name, build_target="install-tests")

Alias("install-tests", env.Install(env["PREFIX"] + "scons/", "scons/scons.py"))
Alias(
    "install-tests",
    env.Install(env["PREFIX"] + "site_scons/", "site_scons/site_init.py"),
)

installFile("SConstruct", build_target="install-tests")

Alias("install-tests", "install")

# Depends('install', DEFAULT_TARGETS)
# Requires('install', DEFAULT_TARGETS)



"""
    PDB2PQR exceptions

    This module represents errors specific to PDB2PQR. Exists mainly to allow
    us to more easily distinguish between code errors and input errors. 

    Parsing utilities provided by Nathan A. Baker (Nathan.Baker@pnl.gov)
    Pacific Northwest National Laboratory

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, 
    are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, 
          this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, 
          this list of conditions and the following disclaimer in the documentation 
          and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
    IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
    OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
    OF THE POSSIBILITY OF SUCH DAMAGE.

"""

__date__ = "20 August 2012"
__author__ = "Kyle Monson"
__version__ = "1.8"

import inspect


class PDB2PQRError(Exception):
    def __init__(self, message):
        self.message = message
        self.line = inspect.currentframe().f_back.f_lineno
        self.filename = inspect.currentframe().f_back.f_code.co_filename

    def __str__(self):
        return """DEBUG INFO: {errorType} {filename}: {line} 
Error encountered: {message}""".format(
            message=self.message,
            errorType=self.__class__.__name__,
            line=self.line,
            filename=self.filename,
        )


class PDBInternalError(PDB2PQRError):
    pass


class PDBInputError(PDB2PQRError):
    pass


class PDB2PKAError(PDB2PQRError):
    pass


class PDBFileParseError(PDB2PQRError):
    def __init__(self, lineno, errorStr):
        self.lineno = lineno
        self.errorStr = errorStr

    def __str__(self):
        return "PDB file parsing error line {lineno}: {errorStr}".format(
            lineno=self.lineno, errorStr=self.errorStr
        )



from fabric.api import *
import tarfile
import os
import zipfile

import fabfile_settings

env.hosts = []

if hasattr(fabfile_settings, "osx_host"):
    osx_host = fabfile_settings.osx_host
    env.hosts.append(osx_host)
    if "@" in osx_host:
        osx_host = osx_host.split("@")[1]
    if ":" in osx_host:
        osx_host = osx_host.split(":")[0]
else:
    osx_host = ""

if hasattr(fabfile_settings, "linux_host"):
    linux_host = fabfile_settings.linux_host
    env.hosts.append(linux_host)
    if "@" in linux_host:
        linux_host = linux_host.split("@")[1]
    if ":" in linux_host:
        linux_host = linux_host.split(":")[0]
else:
    linux_host = ""

if hasattr(fabfile_settings, "run_tests"):
    run_tests = fabfile_settings.run_tests
else:
    run_tests = False

import sys, os

sys.path.append("site_scons")
from defaults import productVersion

pv = productVersion.replace(" ", "_")


class TarWrapper:
    def __init__(self, name, prefix=None):
        self.prefix = prefix
        self.tar = tarfile.open(name, "w:gz")

    def add(self, name, arcname=None):
        if self.prefix is not None:
            if arcname is not None:
                arcname = os.path.join(self.prefix, arcname)
            else:
                arcname = os.path.join(self.prefix, name)

        self.tar.add(name, arcname)
        print ("Packing file: " + name if arcname is None else arcname)

    def close(self):
        self.tar.close()


def pack():
    tar = start_src_tar()
    if env.host == linux_host:
        tar.add("apbs_libs/linux/apbslib.py", "pdb2pka/apbslib.py")
        tar.add("apbs_libs/linux/_apbslib.so", "pdb2pka/_apbslib.so")
    elif env.host == osx_host:
        tar.add("apbs_libs/osx/apbslib.py", "pdb2pka/apbslib.py")
        tar.add("apbs_libs/osx/_apbslib.so", "pdb2pka/_apbslib.so")
    tar.close()


@runs_once
def pack_for_nbcr():
    create_dist_folder()
    tar = start_src_tar("pdb2pqr-src-nbcr-" + pv + ".tar.gz", "pdb2pqr-src-" + pv)
    tar.add("apbs_libs/linux/apbslib.py", "pdb2pka/apbslib.py")
    tar.add("apbs_libs/linux/_apbslib.so", "pdb2pka/_apbslib.so")
    tar.close()
    local("move pdb2pqr-src-nbcr-" + pv + ".tar.gz dist_files\\")


def start_src_tar(name="pdb2pqr.tgz", prefix=None):
    file_list = local("git ls-tree -r --name-only HEAD", capture=True).split("\n")
    tar = TarWrapper(name, prefix)
    for f in file_list:
        tar.add(f)
    return tar


@runs_once
def create_dist_folder():
    with settings(warn_only=True):
        local("mkdir dist_files")


@runs_once
def pack_for_ditro():
    create_dist_folder()
    tar = start_src_tar("pdb2pqr-src-" + pv + ".tar.gz", "pdb2pqr-src-" + pv)
    tar.close()
    local('copy Changelog.md "dist_files\PDB2PQR-' + pv + '-ReleaseNotes.txt"')
    local("move pdb2pqr-src-" + pv + ".tar.gz dist_files\\")


def deploy():
    python = "python2.7"

    put("pdb2pqr.tgz", "~/")
    with settings(warn_only=True):
        run("rm -rf tmp")
        run("mkdir tmp")

    with cd("~/tmp/"):
        run("tar -zxvf ~/pdb2pqr.tgz")
        run(python + " scons/scons.py")

        if run_tests:
            run(python + " scons/scons.py -j 4 pdb2pka-test")
            run(python + " scons/scons.py -j 4 complete-test")


def install_on_deployed():
    python = "python2.7"

    with settings(warn_only=True):
        with cd("~/www/pdb2pqr"):
            run("rm -rf *")

    with cd("~/tmp/"):
        configopts = ""

        if hasattr(fabfile_settings, "APBS"):
            configopts += " APBS=" + fabfile_settings.APBS

        if hasattr(fabfile_settings, "URL"):
            configopts += " URL=" + fabfile_settings.URL

        if hasattr(fabfile_settings, "PREFIX"):
            configopts += " PREFIX=" + fabfile_settings.PREFIX

        if hasattr(fabfile_settings, "OPAL"):
            configopts += " OPAL=" + fabfile_settings.OPAL

        if hasattr(fabfile_settings, "APBS_OPAL"):
            configopts += " APBS_OPAL=" + fabfile_settings.APBS_OPAL

        #         if True:
        #             configopts += ' URL=http://PT24098/d3k084/pdb2pqr_opal'
        #             configopts += ' PREFIX=/Users/d3k084/www/pdb2pqr_opal/'
        #             configopts += ' APBS_OPAL=http://nbcr-222.ucsd.edu/opal2/services/apbs_1.3'
        #             configopts += ' OPAL=http://nbcr-222.ucsd.edu/opal2/services/pdb2pqr_2.0.0'
        #         else:
        #             configopts += ' PREFIX=/Users/d3k084/www/pdb2pqr/'
        #             configopts += ' URL=http://PT24098/d3k084/pdb2pqr'

        run(python + " scons/scons.py " + configopts)
        run(python + " scons/scons.py install " + configopts)


def build_binary_from_deploy():
    create_dist_folder()
    with cd("~/tmp/"):
        os_string = "NOT_SET_FIX_ME"
        if env.host == linux_host:
            os_string = "linux"

        if env.host == osx_host:
            os_string = "osx"

        run("pyinstaller pdb2pqr.spec")

        name = "pdb2pqr-" + os_string + "-bin64-" + pv
        run("mv dist/pdb2pqr dist/" + name)
        with cd("dist/"):
            run("tar -zcvf " + name + ".tar.gz " + name)
            if run_tests:
                with cd(name):
                    run(
                        "./pdb2pqr --ff=parse --verbose --ligand=examples/ligands/LIG_1ABF.mol2 1ABF 1ABF.pqr"
                    )
                    run(
                        "./pdb2pqr --with-ph=7.0 --ph-calc-method=pdb2pka --ff=parse --verbose 1a1p 1a1p.pqr"
                    )

        get("~/tmp/dist/*.tar.gz", "dist_files/")


def linux_bin_cross_platform_test():
    """
    Push the linux bin to a host and test it.
    """
    os_string = "linux"
    name = "pdb2pqr-" + os_string + "-bin64-" + pv
    put("dist_files/" + name + ".tar.gz", "~/")
    run("tar -zxvf ~/" + name + ".tar.gz")
    with cd(name):
        run(
            "./pdb2pqr --ff=parse --verbose --ligand=examples/ligands/LIG_1ABF.mol2 1ABF 1ABF.pqr"
        )
        run(
            "./pdb2pqr --with-ph=7.0 --ph-calc-method=pdb2pka --ff=parse --verbose 1a1p 1a1p.pqr"
        )

    run("rm -rf " + name)


@runs_once
def build_windows():
    local(r"copy apbs_libs\windows\* pdb2pka\\")
    local(r"python scons\scons.py -c")
    local(r"python scons\scons.py")
    if run_tests:
        local(r"python scons\scons.py -j 4 pdb2pka-test")
        local(r"python scons\scons.py -j 7 complete-test")

    build_windows_binary()
    if run_tests:
        test_windows_binary()


@runs_once
def test_windows_binary():
    name = "pdb2pqr-windows-bin64-" + pv
    with lcd(r"dist\\" + name):
        local(
            r"pdb2pqr --ff=parse --verbose --ligand=examples\ligands\LIG_1ABF.mol2 1ABF 1ABF.pqr"
        )
        local(
            "pdb2pqr --with-ph=7.0 --ph-calc-method=pdb2pka --ff=parse --verbose 1a1p 1a1p.pqr"
        )


@runs_once
def build_windows_binary():
    with settings(warn_only=True):
        local("del /Q /F dist")

    local("pyinstaller pdb2pqr.spec")
    name = "pdb2pqr-windows-bin64-" + pv
    local(r"move /Y dist\pdb2pqr dist\\" + name)

    zip_file = zipfile.ZipFile(name + ".zip", "w", zipfile.ZIP_DEFLATED)
    for root, _, files in os.walk(r"dist\\" + name):
        new_root = root.split("/", 1)[-1]
        print root
        for f in files:
            zip_file_path = os.path.join(new_root, f)
            real_file_path = os.path.join(root, f)
            zip_file.write(real_file_path, zip_file_path)
            print "Zipping " + zip_file_path
    zip_file.close()
    create_dist_folder()
    local("move /Y " + name + ".zip" + r" dist_files\\" + name + ".zip")


def build_all_tarballs():
    pack_for_ditro()
    pack_for_nbcr()


def build_all_binaries():
    pack()
    deploy()
    build_binary_from_deploy()
    build_windows()


def deploy_and_install():
    pack()
    deploy()
    install_on_deployed()



"""
    Extentions for PDB2PQR Suite

    This module provides various utilities for the PDB2PQR suite to be
    imported into other Python scripts.
    
    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

    All rights reserved.

    Redistribution and use in source and binary forms, with or without modification, 
    are permitted provided that the following conditions are met:

        * Redistributions of source code must retain the above copyright notice, 
          this list of conditions and the following disclaimer.
        * Redistributions in binary form must reproduce the above copyright notice, 
          this list of conditions and the following disclaimer in the documentation 
          and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
    ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
    WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
    IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
    INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
    BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
    LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
    OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
    OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

import pkgutil

from optparse import OptionGroup, OptionConflictError, Option

_extList = [name for _, name, _ in pkgutil.iter_modules(__path__)]

extDict = {}

for extName in _extList:
    extDict[extName] = __import__(extName,globals(),locals(),[], -1)
    
def setupExtensionsOptions(parser):
    """
    Takes an instance of an OptionParser 
    and adds the options for all extensions
    
    If an extension adds it's own options, those
    options are put in their own group.
    """
    
    if len(extDict) == 0:
        return None
    
    firstGroup = OptionGroup(parser,"Extension options")
    groups = [firstGroup]
    
    for extName, extModule in extDict.items():
        helpArg = {}
        if hasattr(extModule, 'usage'):
            helpArg['help'] = extModule.usage()
            
        extOption = Option('--' + extName, action='append_const', 
                             const = extName, dest = 'active_extensions', **helpArg )
        
        try:
            if hasattr(extModule, 'addExtensionOptions'):
                group = OptionGroup(parser, extName.capitalize() + " extension options")
                group.add_option(extOption)
                
                extModule.addExtensionOptions(group)
                
                if len(group.option_list) == 1:
                    opt = group.option_list[0]
                    group.remove_option(opt.get_opt_string())
                    firstGroup.add_option(opt)
                else:
                    groups.append(group)   
                
            else:
                firstGroup.add_option(extOption)
                
        except OptionConflictError, value:
            print 'Error adding command line options for extension ' + extName + ' ' + '(' + str(value) + ')'
            
    
    for group in groups:
        parser.add_option_group(group)
    
    return groups

class extOutputHelper(object):
    """
    Simple class that makes writing to both file and output simple.
    """
    def __init__(self, routines, outfile):
        self.routines = routines
        self.outfile = outfile

    def write(self, s):
        self.routines.write(s)
        self.outfile.write(s)


"""
    Nucleic Acid Structures for PDB2PQR

    This module contains the base nucleic acid structures for
    pdb2pqr.

    ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------

"""

__date__ = "28 February 2006"
__author__ = "Todd Dolinsky"

import string
from structures import *


class Nucleic(Residue):
    """
    Nucleic class

    This class provides standard features of the nucleic acids listed
    below

    Parameters
        atoms:  A list of Atom objects to be stored in this class
                 (list)
        ref:    The reference object for the amino acid.  Used to
                convert from the alternate naming scheme to the
                main naming scheme.
    """

    def __init__(self, atoms, ref):
        sampleAtom = atoms[-1]

        self.atoms = []
        self.name = sampleAtom.resName
        self.chainID = sampleAtom.chainID
        self.resSeq = sampleAtom.resSeq
        self.iCode = sampleAtom.iCode

        self.ffname = self.name
        self.map = {}
        self.dihedrals = []
        self.patches = []
        self.is3term = 0
        self.is5term = 0
        self.isCterm = 0
        self.isNterm = 0
        self.missing = []
        self.reference = ref

        # Create each atom

        for a in atoms:
            if a.name in ref.altnames:  # Rename atoms
                a.name = ref.altnames[a.name]

            if a.name not in self.map:
                atom = Atom(a, "ATOM", self)
                self.addAtom(atom)

    def createAtom(self, atomname, newcoords):
        """
        Create an atom.  Overrides the generic residue's createAtom().

        Parameters
            atomname:  The name of the atom to add (string)
            newcoords: The coordinates of the atom (list)
        """
        oldatom = self.atoms[0]
        newatom = Atom(oldatom, "ATOM", self)
        newatom.set("x", newcoords[0])
        newatom.set("y", newcoords[1])
        newatom.set("z", newcoords[2])
        newatom.set("name", atomname)
        newatom.set("occupancy", 1.00)
        newatom.set("tempFactor", 0.00)
        newatom.added = 1
        self.addAtom(newatom)

    def addAtom(self, atom):
        """
        Override the existing addAtom - include the link to the
        reference object
        """
        self.atoms.append(atom)
        atomname = atom.get("name")
        self.map[atomname] = atom
        try:
            atom.reference = self.reference.map[atomname]
            for bond in atom.reference.bonds:
                if self.hasAtom(bond):
                    bondatom = self.map[bond]
                    if bondatom not in atom.bonds:
                        atom.bonds.append(bondatom)
                    if atom not in bondatom.bonds:
                        bondatom.bonds.append(atom)
        except KeyError:
            atom.reference = None

    def addDihedralAngle(self, value):
        """
        Add the value to the list of chiangles

        Parameters
            value: The value to be added (float)
        """
        self.dihedrals.append(value)

    def setState(self):
        """
        Adds the termini for all inherited objects
        """
        if self.is5term:
            self.ffname = self.ffname + "5"
        if self.is3term:
            self.ffname = self.ffname + "3"


class ADE(Nucleic):
    """
    Adenosine class

    This class gives data about the Adenosine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "A"

    def setState(self):
        """
        Set the state to distinguish RNA from DNA.
        """
        if self.name == "RA":
            self.ffname = "RA"
        else:
            self.ffname = "DA"
        # if self.hasAtom("O2'"):
        #    self.ffname = "RA"
        # else:
        #    self.ffname = "DA"
        Nucleic.setState(self)


class CYT(Nucleic):
    """
    Cytidine class

    This class gives data about the Cytidine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "C"

    def setState(self):
        """
        Set the state to distinguish RNA from DNA.
        """
        if self.name == "RC":
            self.ffname = "RC"
        else:
            self.ffname = "DC"
        # if self.hasAtom("O2'"):
        #    self.ffname = "RC"
        # else:
        #    self.ffname = "DC"
        Nucleic.setState(self)


class GUA(Nucleic):
    """
    Guanosine class

    This class gives data about the Guanosine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "G"

    def setState(self):
        """
        Set the state to distinguish RNA from DNA.
        """
        if self.name == "RG":
            self.ffname = "RG"
        else:
            self.ffname = "DG"

        # if self.hasAtom("O2'"):
        #    self.ffname = "RG"
        # else:
        #    self.ffname = "DG"
        Nucleic.setState(self)


class THY(Nucleic):
    """
    Thymine class

    This class gives data about the Thymine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "T"

    def setState(self):
        """
        Set the state to distinguish RNA from DNA.  In this case it is
        always DNA.
        """
        self.ffname = "DT"
        Nucleic.setState(self)


class URA(Nucleic):
    """
    Uridine class

    This class gives data about the Uridine object, and inherits
    off the base residue class.
    """

    def __init__(self, atoms, ref):
        """
        Initialize the class

        Parameters
            atoms:      A list of Atom objects to be stored in this class
                        (list)
        """
        Nucleic.__init__(self, atoms, ref)
        self.reference = ref

    def letterCode(self):
        return "U"

    def setState(self):
        """
        Set the state to distinguish RNA from DNA.  In this case it is
        always RNA.
        """
        self.ffname = "RU"
        Nucleic.setState(self)


# Tie the class name to the base name in NA.XML
class RA(ADE):
    pass


class RC(CYT):
    pass


class RG(GUA):
    pass


class DT(THY):
    pass


class RU(URA):
    pass



import argparse

parser = argparse.ArgumentParser(description="Translate a dx file to cube format")
parser.add_argument("dx_input", help="Name of the dx_input file (required arguement)")
parser.add_argument("pqr_input", help="Name of the pqr_input file (required arguement)")
parser.add_argument("output", help="Name of the output file (required arguement)")

args = parser.parse_args()
counter = 0

# DX STUFF
if args.dx_input.endswith(".dx"):
    print "Success"
else:
    print "Error converting file"

try:
    with open(args.dx_input, "r") as in_f, open(args.output, "w") as out_f, open(
        args.pqr_input, "r"
    ) as in_pqr:
        out_f.write(
            "CPMD CUBE FILE.\n" "OUTER LOOP: X, MIDDLE LOOP: Y, INNER LOOP: Z\n"
        )

        # Discard comments at top of file.
        line = in_f.readline()
        newline = in_pqr.readline()
        while line.startswith("#"):
            line = in_f.readline()
        while newline.startswith("REMARK"):
            newline = in_pqr.readline()

        split_line = line.split()
        grid_sizes = [int(x) * -1 for x in split_line[-3:]]

        split_line = in_f.readline().split()

        origin = [float(x) for x in split_line[-3:]]

        parameter_fmt = "{:>4} {:>11.6f} {:>11.6f} {:>11.6f}\n"
        try:
            while newline.startswith("ATOM"):
                pqr_line = in_pqr.readline()
                new_split_line = pqr_line.split()
                atom_num = new_split_line[1]
        except IndexError:
            pass
        in_pqr.seek(0)
        newline = in_pqr.readline()
        while newline.startswith("REMARK"):
            newline = in_pqr.readline()

        origin_line = parameter_fmt.format(atom_num, *origin)
        out_f.write(origin_line)

        for x in xrange(3):
            split_line = in_f.readline().split()
            grid_dims = [float(item) for item in split_line[-3:]]

            dim_lin = parameter_fmt.format(grid_sizes[x], *grid_dims)
            out_f.write(dim_lin)

        atoms_parameter_fmt = "{:>4} {:>11.6f} {:>11.6f} {:>11.6f} {:>11.6f}\n"
        a = True
        xreal_center = []
        yreal_center = []
        zreal_center = []
        try:
            while a == True:
                if not newline.startswith("TER"):
                    new_split_line = newline.split()
                    radius = new_split_line[-1]
                    xyz = new_split_line[-5:-2]
                    line_atom_num = new_split_line[1]
                    atom_radius = new_split_line[-1]
                    pqr_lin = atoms_parameter_fmt.format(
                        int(line_atom_num),
                        float(new_split_line[-2]),
                        float(xyz[0]),
                        float(xyz[1]),
                        float(xyz[2]),
                    )
                    out_f.write(pqr_lin)
                    newline = in_pqr.readline()
                    xreal_center.append(float(xyz[0]))
                    yreal_center.append(float(xyz[1]))
                    zreal_center.append(float(xyz[2]))
                else:
                    a = False
        except IndexError:
            pass

        x_avg = sum(xreal_center) / float(atom_num)
        y_avg = sum(yreal_center) / float(atom_num)
        z_avg = sum(zreal_center) / float(atom_num)
        print x_avg, y_avg, z_avg

        # print origin
        # new_origin = []
        # for item in origin:
        #    newitem = item/0.529177
        # new_new = item/2 + newitem/2
        #    new_origin.append(newitem)
        # print new_origin

        # Consume unneeded object lines.
        in_f.readline()
        in_f.readline()

        ##TODO: put atoms here

        value_format = ["{:< 13.5E}"]
        value_format = " ".join(value_format * 6) + "\n"
        print value_format
        group = []
        line = in_f.readline()
        while not line.startswith("attribute"):
            values = [float(item) for item in line.split()]
            group.extend(values)

            if len(group) >= 6:
                out_f.write(value_format.format(*group))
                group = []

            line = in_f.readline()

        if group:
            group_strs = ["{:< 13.5E}".format(item) for item in group]
            out_f.write(" ".join(group_strs))


except IOError:
    print "file doesn't exist"
"""
for line in dx_data:
    dx_data_2 = dx_data
    data_point = line.strip("\n")
    dx_data_2.pop(0)
    dx_data._2append(data_point)
    

for line in dx_data_2:
    if line.startswith('#'):
        dx_data_2.remove(line)
        continue
    elif "gridpositions" in line:
        new_line = line.split()
        grid_values = new_line[-3:]
        #grid values for cube file, will need grid_valies[0]
        break
    elif line.startswith('delta'):
        line.insert('0')
        line.remove(delta)



#PQR STUFF
if args.pqr_input.endswith('.pqr'):
    print "yay"
else:
    print "nay"

try:
    with open(args.pqr_input, 'r') as f:
        pqr_data = f.readlines()
        
except IOError:
    print "file doesn't exist"
    
for line in pqr_data:
    data_point = line.strip("\n")
    pqr_data.pop(0)
    pqr_data.append(data_point)

for line in pqr_data:
    if line.startswith('REMARK'):
        pqr_data.remove(line)
        continue
    counter +=1
    #the number next to the origin in cube format
"""



import os
import subprocess
from pdbmender.formats import read_pdb_line, new_pdb_line
from pdbmender.constants import (
    PROTEIN_RESIDUES,
    TITRATABLE_RESIDUES,
    RESIDUE_REFSTATE,
    RENAME_ATOMS,
)

SCRIPT_DIR = os.path.dirname(os.path.abspath(__file__))


def correct_insertion_codes(pdb_in, pdb_out, insertion_offset=3000):
    with open(pdb_in) as f:
        lines = f.readlines()

    ninsertions = -1
    inserted_block = None
    renumbered_res = {}
    new_lines = ""
    for line in lines:
        if line.startswith("ATOM "):
            aname, anumb, resname, chain, resnumb, x, y, z = read_pdb_line(line)

            if chain == " ":
                chain = "_"

            insertion_code = line[26]
            if insertion_code.strip():
                new_res_trigger = False
                if inserted_block != insertion_code:
                    ninsertions += 1
                    new_res_trigger = True

                inserted_block = insertion_code
                resnumb_offset = insertion_offset + ninsertions
                resnumb += resnumb_offset

                if new_res_trigger:
                    while (
                        chain in renumbered_res.keys()
                        and resnumb in renumbered_res[chain].keys()
                    ):
                        resnumb += 1
                        ninsertions += 1
                    if resnumb not in renumbered_res.keys():
                        if chain not in renumbered_res:
                            renumbered_res[chain] = {}
                        renumbered_res[chain][resnumb] = (
                            resnumb - resnumb_offset,
                            insertion_code,
                        )
            else:
                ninsertions = -1

            new_line = new_pdb_line(
                anumb,
                aname,
                resname,
                resnumb,
                x,
                y,
                z,
                chain=chain,
            )
            new_lines += new_line

    with open(pdb_out, "w") as f:
        f.write(new_lines)

    return renumbered_res


def mend_pdb(
    pdb_to_clean, pdb_cleaned, ff, ffout, logfile="LOG_pdb2pqr", hopt=True, debug=False
):
    pdb_fixed = pdb_to_clean.replace(".pdb", "_fixed.pdb")
    renumbered_res = correct_insertion_codes(pdb_to_clean, pdb_fixed)
    try:
        # TODO: Port pdb2pqr to py3 and import it as a module
        cmd = (
            "python2 {0}/pdb2pqr/pdb2pqr.py {1} {2} "
            "--ff {3} --ffout {4} --drop-water -v --chain {6} > {5} 2>&1 ".format(
                SCRIPT_DIR,
                pdb_fixed,
                pdb_cleaned,
                ff,
                ffout,
                logfile,
                "" if hopt else "--noopt",
            )
        )
        subprocess.run(
            cmd,
            shell=True,
            check=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
    except subprocess.CalledProcessError as e:
        raise Exception(
            "pdb2pqr did not run successfully\nMessage: {}".format(
                e.stderr.decode("ascii")
            )
        )
    if not debug:
        os.system("rm -f {}".format(pdb_fixed))
    return renumbered_res


def add_tautomers(
    pdb_in,
    chains_res,
    ff_family,
    outputpqr,
    logfile="LOG_addHtaut",
    pdb_prep="addhtaut_cleaned.pdb",
    to_exclude=[],
    terminal_offset=5000,
):
    removed_pdb = prepare_for_addHtaut(
        pdb_in, pdb_prep, chains_res, to_exclude, terminal_offset=terminal_offset
    )
    sites_addHtaut = select_tautomer_sites(chains_res)
    if len(sites_addHtaut.strip()) == 0:
        os.system("cp {} {}".format(pdb_prep, outputpqr))
    else:
        try:
            # TODO rewrite addHtaut as python module
            cmd = "{}/addHtaut {} {} {} > {}".format(
                SCRIPT_DIR,
                pdb_prep,
                ff_family,
                sites_addHtaut,
                outputpqr,
                logfile,
            )
            subprocess.run(
                cmd,
                shell=True,
                check=True,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            )
        except subprocess.CalledProcessError as e:
            raise Exception(
                "addHtaut did not run successfully\nMessage: {}".format(
                    e.stderr.decode("ascii")
                )
            )
    return removed_pdb


def correct_names(resnumb, resname, aname, titrating_sites, termini):
    # TODO: some of these are no longer used as it is done by pdb2pqr
    def change_aname(aname, restype):
        not_correct_names = list(restype.keys())
        for not_corrected in not_correct_names:
            if aname == not_corrected:
                aname = restype[not_corrected]
        return aname

    NTR_numb, CTR_numb = termini
    restype = None
    if resnumb == CTR_numb:
        restype = "CTR"

    elif resnumb == NTR_numb:
        restype = "NTR"

    if restype and restype in list(RENAME_ATOMS.keys()):
        aname = change_aname(aname, RENAME_ATOMS[restype])

    if resnumb in titrating_sites:
        if resname not in PROTEIN_RESIDUES:
            for tit_res in TITRATABLE_RESIDUES:
                if tit_res[:2] == resname[:2]:
                    restype, resname = tit_res, tit_res

        if resname in list(RENAME_ATOMS.keys()):
            aname = change_aname(aname, RENAME_ATOMS[resname])

    if resname in list(RESIDUE_REFSTATE.keys()):
        resname = RESIDUE_REFSTATE[resname]

    return aname, resname


def prepare_for_addHtaut(pdb_in, pdb_out, chains_res, to_exclude, terminal_offset=5000):
    with open(pdb_in) as f:
        content = f.readlines()

    sites = {}
    termini = {}
    for chain, residues in chains_res.items():
        sites[chain] = []
        termini[chain] = [None, None]
        for resnumb, resname in residues.items():
            if resname in ("NTR", "CTR"):
                resnumb = int(resnumb)
                if resname == "NTR":
                    termini[chain][0] = resnumb
                elif resname == "CTR":
                    termini[chain][1] = resnumb
                resnumb += terminal_offset
            sites[chain].append(resnumb)

    new_pdb_text = ""
    removed_pdb_lines = []
    resnumb_max = 0
    chains = sites.keys()
    for line in content:
        if line.startswith("ATOM"):
            termini_trigger = False
            aname, anumb, resname, chain, resnumb, x, y, z = read_pdb_line(line)

            if chain in chains:
                resnumb_max = resnumb
                sites_numbs = sites[chain]
                aname, resname = correct_names(
                    resnumb, resname, aname, sites_numbs, termini[chain]
                )

                if resnumb in termini[chain]:
                    termini_trigger = True

            if resname == "CYS":
                change_atoms = {"1CB": "CB", "1SG": "SG"}
                if aname in change_atoms.keys():
                    aname = change_atoms[aname]
            if (
                aname in ("O1", "O2", "OT1", "OT2", "H1", "H2", "H3")
                and not termini_trigger
                and resname in PROTEIN_RESIDUES
            ):
                if aname == "O1":
                    aname = "O"
                elif aname == "H1":
                    aname = "H"
                else:
                    continue

            new_line = new_pdb_line(
                anumb,
                aname,
                resname,
                resnumb,
                x,
                y,
                z,
                chain=chain,
            )
            if chain in chains:
                new_pdb_text += new_line
            elif (
                aname not in ("O1", "O2", "OT1", "OT2", "H1", "H2", "H3")
                or resname in to_exclude
            ):
                removed_pdb_lines.append(new_line)

    with open(pdb_out, "w") as f_new:
        f_new.write(new_pdb_text)

    resnumb_old = resnumb_max + 1
    removed_pdb_text = ""
    for line in removed_pdb_lines:
        (aname, anumb_old, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
        anumb += 1
        resnumb += resnumb_old
        while resnumb < resnumb_max:
            resnumb += resnumb_old
        removed_pdb_text += new_pdb_line(
            anumb, aname, resname, resnumb, x, y, z, chain=chain
        )
        resnumb_max = resnumb

    with open("removed.pqr", "w") as f_new:
        f_new.write(removed_pdb_text)
    return removed_pdb_text


def get_cys_bridges(f_pdb2pqr_log):
    CYS_bridges = {}
    with open(f_pdb2pqr_log) as f:
        trigger_error = ""
        for line in f:
            if "patched with CYX" in line:
                parts = line.split("patched")[0].replace("PATCH INFO: ", "").split()
                resname, chain, resnumb = parts

                if not chain in CYS_bridges:
                    CYS_bridges[chain] = []
                cys_res_numb = int(parts[-1])
                if cys_res_numb not in CYS_bridges[chain]:
                    CYS_bridges[chain].append(cys_res_numb)
            if (
                "error" in line.lower()
                and "Error parsing line: invalid literal for int() with base 10"
                not in line
            ):
                trigger_error += line
        if trigger_error:
            raise Exception(
                "Found errors while parsing the input structure to PDB2PQR:\n"
                + trigger_error
            )
    return CYS_bridges


def identify_cter(inputpqr, cur_ctrs):
    new_ctrs = {}
    with open(inputpqr) as f:
        for line in f:
            if line.startswith("ATOM "):
                (aname, _, resname, chain, resnumb, _, _, _) = read_pdb_line(line)
                if (
                    aname
                    in (
                        "CT",
                        "OT",
                        "OT1",
                        "OT2",
                        "O1",
                        "O2",
                        "OXT",
                    )
                    and chain in cur_ctrs
                    and resnumb != cur_ctrs[chain]
                    and resname in [*TITRATABLE_RESIDUES, *PROTEIN_RESIDUES, "CY0"]
                ):
                    new_ctrs[chain] = resnumb

    return new_ctrs


def identify_tit_sites(f_in, chains, nomenclature="PDB", add_ser_thr=False):
    def add_site(chain_res, chain, resnumb, resname):
        if resname in ("NTR", "CTR"):
            chain_res[chain][str(resnumb)] = resname
        else:
            chain_res[chain][resnumb] = resname
        return chain_res

    chain_res = {chain: {} for chain in chains}
    skip_NTR = {chain: False for chain in chains}
    last_res = None
    with open(f_in) as f:
        content = f.readlines()

    nline = 0
    resnumb = None
    resname = None
    for line in content:
        nline += 1
        if line.startswith("ATOM "):
            (aname, anumb, resname, chain, resnumb, x, y, z) = read_pdb_line(line)
            insertion_code = line[26].strip()
            if insertion_code:
                print("NOT CONSIDERING:", line)
                continue

            last_res = resnumb
            last_chain = chain
            chain_sites = []

            if chain in chains:
                chain_sites = chain_res[chain]

                if nomenclature == "CHARMM":
                    for can_res, charmm_res in CHARMM_protomers.items():
                        if resname in charmm_res:
                            resname = can_res

                if resname in PROTEIN_RESIDUES or resname in TITRATABLE_RESIDUES:
                    if resnumb not in chain_sites:
                        if not chain_res[chain] and not skip_NTR[chain]:
                            if resname == "PRO":
                                skip_NTR[chain] = True
                                continue
                            chain_res = add_site(chain_res, chain, resnumb, "NTR")

                        if (
                            resname in TITRATABLE_RESIDUES
                            and resname != "NTR"
                            and resname != "CTR"
                        ) and (resname not in ("SER", "THR") or add_ser_thr):
                            chain_res = add_site(chain_res, chain, resnumb, resname)

                    if "CTR" not in chain_sites and aname in (
                        "CT",
                        "OT",
                        "OT1",
                        "OT2",
                        "O1",
                        "O2",
                        "OXT",
                    ):
                        chain_res = add_site(chain_res, chain, resnumb, "CTR")

    if "CTR" not in chain_sites and aname in (
        "CT",
        "OT",
        "OT1",
        "OT2",
        "O1",
        "O2",
        "OXT",
    ):
        chain_res = add_site(chain_res, chain, last_res, "CTR")

    return chain_res


def select_tautomer_sites(chains_res):
    sites_addHtaut = ""
    for chain in chains_res.keys():
        for res in chains_res[chain]:
            if chains_res[chain][res] == "NTR":
                continue
            sites_addHtaut += "{0}--{1}--{2},".format(
                res, chains_res[chain][res], chain.replace(" ", "_")
            )

    if len(sites_addHtaut) > 0 and sites_addHtaut[-1] == ",":
        sites_addHtaut = sites_addHtaut[:-1]
    return sites_addHtaut


def rm_cys_bridges(chains_res, logfile_mend):
    cys_bridges = get_cys_bridges(logfile_mend)
    final_chains_res = {chain: {} for chain in chains_res}
    for chain in final_chains_res:
        for res in chains_res[chain]:
            if chain in cys_bridges and res in cys_bridges[chain]:
                continue
            final_chains_res[chain][res] = chains_res[chain][res]

    return final_chains_res, cys_bridges



""" inputgen class

    Create an APBS input file using psize data

    Written by Todd Dolinsky based on original sed script by Nathan Baker

        ----------------------------
   
    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin; 
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific 
    Northwest National Laboratory, operated by Battelle Memorial Institute, 
    Pacific Northwest Division for the U.S. Department Energy.; 
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification, 
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice, 
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice, 
		  this list of conditions and the following disclaimer in the documentation 
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------
"""

# User - Definable Variables: Default values

# cfac = 1.7                  # Factor by which to expand mol dims to
                              # get coarse grid dims
# fadd = 20                   # Amount to add to mol dims to get fine
                              # grid dims
# space = 0.50                # Desired fine mesh resolution
# gmemfac = 200               # Number of bytes per grid point required 
                              # for sequential MG calculation 
# gmemceil = 400              # Max MB allowed for sequential MG
                              # calculation.  Adjust this to force the
                              # script to perform faster calculations (which
                              # require more parallelism).
# ofrac = 0.1                  # Overlap factor between mesh partitions
# redfac = 0.25               # The maximum factor by which a domain
                              # dimension can be reduced during focusing
__date__ = "21 April 2007"
__author__ = "Todd Dolinsky, Nathan Baker, Yong Huang"

import string, sys
import psize
import pickle
import os.path
import utilities

class Elec:
    """
        An object for the ELEC section of an APBS input file
    """
    def __init__(self, pqrpath, size, method, asyncflag, istrng=0, potdx=False):
        """
            Initialize the variables that can be set in this object
            Users can modify any of these variables (that's why
            they're here!)
        """

        # If this is an async or parallel calc, we want to use
        # the per-grid dime rather than the global dime.
        
        self.dime = size.getFineGridPoints()
        gmem = 200.0 * self.dime[0] * self.dime[1] * self.dime[2] / 1024.0 / 1024.0
        if method == "": # method not named - use ceiling
            if gmem > size.getConstant("gmemceil"): method = "mg-para"
            else: method = "mg-auto"

        if method == "mg-para":
            self.dime = size.getSmallest()

        self.method = method
        self.istrng = istrng
        self.glen = size.getCoarseGridDims()
        self.cglen = size.getCoarseGridDims()
        self.fglen = size.getFineGridDims()
        self.pdime = size.getProcGrid()
        
        self.label = ""
        self.nlev = 4
        self.ofrac = 0.1
        self.async = 0
        self.asyncflag = asyncflag
        self.cgcent = "mol 1"
        self.fgcent = "mol 1"
        self.gcent = "mol 1"
        self.mol = 1
        self.lpbe = 1
        self.npbe = 0
        self.bcfl = "sdh"
        self.ion = [[-1,1.815],[1,1.875]] # Multiple ions possible
        self.pdie = 2.0
        self.sdie = 78.54
        self.srfm = "smol"
        self.chgm = "spl2"
        self.sdens = 10.0
        self.srad = 1.4
        self.swin = 0.3
        self.temp = 298.15
        self.gamma = 0.105
        self.calcenergy = "total"
        self.calcforce = "no"
        if potdx:
            self.write = [["pot", "dx", pqrpath]]
        else:
            self.write = [["pot", "dx", "pot"]] # Multiple write statements possible
    
    def __str__(self):
        """
            Return the elec statement as a string. Check the method
            to see which keywords to use.
        """
        text = "elec %s\n" % self.label
        text += "    %s\n" % self.method
        text += "    dime %i %i %i\n" % (self.dime[0], self.dime[1], self.dime[2])
        if self.method == "mg-manual":
            text += "    glen %.3f %.3f %.3f\n" % (self.glen[0], self.glen[1], self.glen[2])
            text += "    gcent %s\n" % self.gcent
        elif self.method == "mg-auto":
            text += "    cglen %.4f %.4f %.4f\n" % (self.cglen[0], self.cglen[1], self.cglen[2])
            text += "    fglen %.4f %.4f %.4f\n" % (self.fglen[0], self.fglen[1], self.fglen[2])
            text += "    cgcent %s\n" % self.cgcent
            text += "    fgcent %s\n" % self.fgcent
        elif self.method == "mg-para":
            text += "    pdime %i %i %i\n" % (self.pdime[0], self.pdime[1], self.pdime[2])
            text += "    ofrac %.1f\n" % self.ofrac
            text += "    cglen %.4f %.4f %.4f\n" % (self.cglen[0], self.cglen[1], self.cglen[2])
            text += "    fglen %.4f %.4f %.4f\n" % (self.fglen[0], self.fglen[1], self.fglen[2])
            text += "    cgcent %s\n" % self.cgcent
            text += "    fgcent %s\n" % self.fgcent
            if self.asyncflag == 1:
                text += "    async %i\n" % self.async
        text += "    mol %i\n" % self.mol
        if self.lpbe: text += "    lpbe\n"
        else: text += "    npbe\n"
        text += "    bcfl %s\n" % self.bcfl
        if self.istrng > 0:
            for ion in self.ion:
                text += "    ion charge %.2f conc %.3f radius %.4f\n" % (ion[0], self.istrng, ion[1])               
        text += "    pdie %.4f\n" % self.pdie                
        text += "    sdie %.4f\n" % self.sdie                
        text += "    srfm %s\n" % self.srfm                   
        text += "    chgm %s\n" % self.chgm
        text += "    sdens %.2f\n" % self.sdens
        text += "    srad %.2f\n" % self.srad          
        text += "    swin %.2f\n" % self.swin         
        text += "    temp %.2f\n" % self.temp     
        text += "    calcenergy %s\n" % self.calcenergy
        text += "    calcforce %s\n" % self.calcforce
        for write in self.write:
            text += "    write %s %s %s\n" % (write[0], write[1], write[2])
        text += "end\n"
        return text
        
class Input:
    """
        The input class.  Each input object is one APBS input file.
    """

    def __init__(self, pqrpath, size, method, asyncflag, istrng=0, potdx=False):
        """
            Initialize the input file class.  Each input file contains
            a PQR name, a list of elec objects, and a list of strings
            containing print statements.  For starters assume two
            ELEC statements are needed, one for the inhomgenous and
            the other for the homogenous dielectric calculations.

            Users can edit the elec statements and the print statements.

            This assumes you have already run psize, either by
                 size.runPsize(/path/to/pqr) or

                 size.parseString(string)
                 size.setAll()

            Parameters
                pqrpath:   The path to the PQR file (string)
                size:      The Psize object (psize)
                method:    The method (para, auto, manual, async) to use
                asyncflag: 1 if async is desired, 0 otherwise
        """ 

        self.pqrpath = pqrpath
        self.asyncflag = asyncflag

        # Initialize variables to default elec values

        elec1 = Elec(pqrpath, size, method, asyncflag, istrng, potdx)
        if not potdx:
            elec2 = Elec(pqrpath, size, method, asyncflag, istrng, potdx)
            setattr(elec2, "sdie", 2.0)
            setattr(elec2, "write", [])
        else:
            elec2 = ""
        self.elecs = [elec1, elec2]
     
        self.pqrname = os.path.basename(pqrpath)

        if not potdx:
            self.prints = ["print elecEnergy 2 - 1 end"]     
        else:
            self.prints = ["print elecEnergy 1 end"]

    def __str__(self):
        """
            Return the text of the input file
        """
        text  = "read\n"
        text += "    mol pqr %s\n" % self.pqrname
        text += "end\n"
        for elec in self.elecs:
            text += str(elec)            
        for prints in self.prints:
            text += prints
        text += "\nquit\n"
        return text
  
    def printInputFiles(self):
        """
            Make the input file(s) associated with this object
        """
        base_pqr_name = utilities.getPQRBaseFileName(self.pqrpath)
        if self.asyncflag == 1:
            outname = base_pqr_name + "-para.in"

            # Temporarily disable async flag
            for elec in self.elecs:
                elec.asyncflag = 0
            file = open(outname, "w")
            file.write(str(self))
            file.close()

            # Now make the async files
            elec = self.elecs[0]
            
            nproc = elec.pdime[0] * elec.pdime[1] * elec.pdime[2]
            for i in range(int(nproc)):
                outname = base_pqr_name + "-PE%i.in" % i
                for elec in self.elecs:
                    elec.asyncflag = 1
                    elec.async = i
                file = open(outname, "w")
                file.write(str(self))
                file.close()
        
        else:
            outname = base_pqr_name + ".in"
            file = open(outname, "w")
            file.write(str(self))
            file.close()

    def dumpPickle(self):
        """
            Make a Python pickle associated with the APBS input parameters
        """
        base_pqr_name = utilities.getPQRBaseFileName(self.pqrpath)
        outname = base_pqr_name + "-input.p"
        pfile = open(outname, "w")
        pickle.dump(self, pfile)
        pfile.close()
        

def splitInput(filename):
    """
        Split the parallel input file into multiple async file names

        Parameters
            filename:  The path to the original parallel input
                       file (string)
    """
    nproc = 0
    file = open(filename, 'rU')
    text = ""
    while 1:
        line = file.readline()
        if line == "": break
        text += line
        line = string.strip(line)
        if line.startswith("pdime"): # Get # Procs
            words = string.split(line)
            nproc = int(words[1]) * int(words[2]) * int(words[3])

    if nproc == 0:
        sys.stderr.write("%s is not a valid APBS parallel input file!\n" % filename)
        sys.stderr.write("The inputgen script was unable to asynchronize this file!\n")
        sys.exit(2)

    base_pqr_name = utilities.getPQRBaseFileName(filename)
    for i in range(nproc):
        outname = base_pqr_name + "-PE%i.in" % i
        outtext = string.replace(text, "mg-para\n","mg-para\n    async %i\n" % i)
        outfile = open(outname, "w")
        outfile.write(outtext)
        outfile.close()
          
def usage():
    """
        Display the usage information for this script
    """
    size = psize.Psize()
    usage = "\n"
    usage = usage + "Use this script to generate new APBS input files or split an existing\n"
    usage = usage + "parallel input file into multiple async files.\n\n"
    usage = usage + "Usage: inputgen.py [opts] <filename>\n"
    usage = usage + "Optional Arguments:\n"
    usage = usage + "  --help               : Display this text\n"
    usage = usage + "  --split              : Split an existing parallel input file to multiple\n"
    usage = usage + "                         async input files.\n"
    usage = usage + "  --potdx              : Create an input to compute an electrostatic potential map.\n"
    usage = usage + "  --method=<value>     : Force output file to write a specific APBS ELEC\n"
    usage = usage + "                         method.  Options are para (parallel), auto\n"
    usage = usage + "                         (automatic), manual (manual), or async (asynchronous).\n"
    usage = usage + "                         solve.  async will result in multiple input files.\n"
    usage = usage + "  --cfac=<value>       : Factor by which to expand molecular dimensions to\n"
    usage = usage + "                         get coarse grid dimensions.\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("cfac")
    usage = usage + "  --fadd=<value>       : Amount to add to molecular dimensions to get fine\n"
    usage = usage + "                         grid dimensions.\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("fadd")
    usage = usage + "  --space=<value>      : Desired fine mesh resolution\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("space")
    usage = usage + "  --gmemfac=<value>    : Number of bytes per grid point required\n"
    usage = usage + "                         for sequential MG calculation\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("gmemfac")
    usage = usage + "  --gmemceil=<value>   : Max MB allowed for sequential MG\n"
    usage = usage + "                         calculation.  Adjust this to force the\n"
    usage = usage + "                         script to perform faster calculations (which\n"
    usage = usage + "                         require more parallelism).\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("gmemceil")
    usage = usage + "  --ofrac=<value>       : Overlap factor between mesh partitions\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("ofrac")
    usage = usage + "  --redfac=<value>     : The maximum factor by which a domain\n"
    usage = usage + "                         dimension can be reduced during focusing\n"
    usage = usage + "                         [default = %g]\n" % size.getConstant("redfac")
    usage = usage + "  --istrng=<value>     : Ionic strength (M). Na+ anc Cl- ions will be used\n"
    sys.stderr.write(usage)
    sys.exit(2)

def main():

    import getopt
    filename = ""
    shortOptList = ""
    longOptList = ["help","split","potdx","method=","cfac=","space=","gmemceil=","gmemfac=","ofrac=","redfac=","istrng="]

    try:
        opts, args = getopt.getopt(sys.argv[1:], shortOptList, longOptList)
    except getopt.GetoptError, details:
        sys.stderr.write("Option error (%s)!\n" % details)
        usage()
        
    if len(args) != 1:
        sys.stderr.write("Invalid argument list!\n")
        usage()
    else:
        filename = args[0]

    method = ""
    size = psize.Psize()
    async = 0
    split = 0
    istrng = 0
    potdx = 0
    
    for o, a in opts:
        if o == "--help":
            usage()
        if o == "--split": split = 1
        if o == "--potdx": potdx = 1
        if o == "--method":
            if a == "para":
                sys.stdout.write("Forcing a parallel calculation\n")
                method = "mg-para"
            elif a == "auto":
                sys.stdout.write("Forcing a sequential calculation\n")
                method = "mg-auto"
            elif a == "async":
                sys.stdout.write("Forcing an asynchronous calculation\n")
                method = "mg-para"
                async = 1
            elif a == "manual":
                sys.stdout.write("Forcing a manual calculation\n")
                method = "mg-manual"
            else:
                sys.stdout.write("Incorrect method argument: %s\n" % a)
                sys.stdout.write("Defaulting to memory dependent result\n")
        if o == "--cfac":
            size.setConstant("cfac", float(a))
        if o == "--space":
            size.setConstant("space", float(a))
        if o == "--gmemfac":
            size.setConstant("gmemfac", int(a))
        if o == "--gmemceil":
            size.setConstant("gmemceil",  int(a))
        if o == "--ofrac":
            size.setConstant("ofrac", float(a))
        if o == "--redfac":
            size.setConstant("redfac", float(a))
        if o == "--istrng":
            istrng = float(a)

    if split == 1:
        splitInput(filename)
    else:
        size.runPsize(filename)
        input = Input(filename, size, method, async, istrng, potdx)
        input.printInputFiles()

if __name__ == "__main__": main()



"""
    Routines for PDB2PQR

    This module contains the protein object used in PDB2PQR and methods
    used to correct, analyze, and optimize that protein.

    ----------------------------

    PDB2PQR -- An automated pipeline for the setup, execution, and analysis of
    Poisson-Boltzmann electrostatics calculations

    Copyright (c) 2002-2011, Jens Erik Nielsen, University College Dublin;
    Nathan A. Baker, Battelle Memorial Institute, Developed at the Pacific
    Northwest National Laboratory, operated by Battelle Memorial Institute,
    Pacific Northwest Division for the U.S. Department Energy.;
    Paul Czodrowski & Gerhard Klebe, University of Marburg.

	All rights reserved.

	Redistribution and use in source and binary forms, with or without modification,
	are permitted provided that the following conditions are met:

		* Redistributions of source code must retain the above copyright notice,
		  this list of conditions and the following disclaimer.
		* Redistributions in binary form must reproduce the above copyright notice,
		  this list of conditions and the following disclaimer in the documentation
		  and/or other materials provided with the distribution.
        * Neither the names of University College Dublin, Battelle Memorial Institute,
          Pacific Northwest National Laboratory, US Department of Energy, or University
          of Marburg nor the names of its contributors may be used to endorse or promote
          products derived from this software without specific prior written permission.

	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
	ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
	IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
	INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
	BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
	DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
	LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
	OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
	OF THE POSSIBILITY OF SUCH DAMAGE.

    ----------------------------

"""

__date__ = "1 August 2008"
__author__ = "Jens Erik Nielsen, Todd Dolinsky, Yong Huang"

CELL_SIZE = 2
BUMP_DIST = 2.0
BUMP_HDIST = 1.5
BUMP_HYDROGEN_SIZE = 0.5
BUMP_HEAVY_SIZE = 1.0
BONDED_SS_LIMIT = 2.5
PEPTIDE_DIST = 1.7
REPAIR_LIMIT = 10
AAS = ["ALA", "ARG", "ASH", "ASN", "ASP", "CYS", "CYM", "GLN", "GLU", "GLH", "GLY", \
       "HIS", "HID", "HIE", "HIP", "HSD", "HSE", "HSP", "ILE", "LEU", "LYS", "LYN", \
       "MET", "PHE", "PRO", "SER", "THR", "TRP", "TYR", "TYM", "VAL"]
NAS = ["A", "A5", "A3", "C", "C5", "C3", "G", "G5", "G3", "T", "T5", "T3", "U", \
       "U5", "U3", "RA", "RG", "RC", "RU", "DA", "DG", "DC", "DT"]

import math
import copy
from pdb import *
from utilities import *
from quatfit import *
from forcefield import *
from structures import *
from protein import *
from definitions import *
from StringIO import StringIO
from errors import PDBInputError, PDBInternalError, PDB2PKAError
from pprint import pformat


class Routines:
    def __init__(self, protein, verbose, definition=None):
        """
            Initialize the Routines class.  The class contains most
            of the main routines that run PDB2PQR

            Parameters
                protein:  The protein to run PDB2PQR on (Protein)
                verbose:  A flag to determine whether to write to
                          stdout
        """
        self.protein = protein
        self.definition = definition
        self.aadef = None
        self.verbose = verbose
        self.warnings = []
        self.cells = {}
        if definition != None:
            self.aadef = definition.getAA()
            self.nadef = definition.getNA()


    def write(self, message, indent=0):
        """
            Write a message to stdout for debugging if verbose

            Parameters
                message: The message to write (string)
                indent : The indent level (int, default=0)
        """
        out = ""
        if self.verbose:
            for i in range(indent):
                out += "\t"
            out += message
            sys.stdout.write(out)

    def getWarnings(self):
        """
            Get all warnings generated from routines
        """
        return self.warnings

    def applyNameScheme(self, forcefield):
        """
            Apply the naming scheme of the given forcefield to the atoms
            within the protein

            Parameters
                forcefield: The forcefield object (forcefield)

        """
        self.write("Applying the naming scheme to the protein...")
        for residue in self.protein.getResidues():
            if isinstance(residue, (Amino, WAT, Nucleic)):
                resname = residue.ffname
            else:
                resname = residue.name

            residue.blockname = None
            for atom in residue.getAtoms():
                rname, aname = forcefield.getNames(resname, atom.name)

                print(residue.name, atom.name, rname, aname)

                if resname not in ['LIG', 'WAT', 'ACE', 'NME'] and rname != None:
                    try:
                        if residue.isNterm and residue.name == 'PRO' and rname == 'NT3' and aname in ('H1', 'H2'):
                            rname = 'DELETE'
                            aname = None

                        elif (residue.isNterm or residue.isCterm) and rname != residue.name:                            
                            if residue.blockname:
                                rname = residue.blockname
                            elif rname == 'CY0':
                                residue.blockname = rname
                            else:
                                rname = residue.name
                    except AttributeError:
                        pass
                else:
                    rname = 'DELETE'

                if aname != None and rname != None:
                    atom.resName = rname
                    atom.name = aname
                elif rname == 'DELETE':
                    atom.resName = None
                    atom.name = None


        self.write("Done.\n")

    def applyForcefield(self, forcefield):
        """
            Apply the forcefield to the atoms within the protein

            Parameters
                forcefield: The forcefield object (forcefield)
            Returns
                hitlist:    A list of atoms that were found in
                            the forcefield (list)
                misslist:   A list of atoms that were not found in
                            the forcefield (list)
        """
        self.write("Applying the forcefield to the protein...")
        misslist = []
        hitlist = []
        for residue in self.protein.getResidues():
            if isinstance(residue, (Amino, WAT, Nucleic)):
                resname = residue.ffname
            else:
                resname = residue.name

            # Apply the parameters

            for atom in residue.getAtoms():
                atomname = atom.get("name")
                charge, radius = forcefield.getParams(resname, atomname)

                if charge != None and radius != None:
                    atom.set("ffcharge", charge)
                    atom.set("radius", radius)
                    hitlist.append(atom)
                else:
                    misslist.append(atom)

        self.write("Done.\n")
        return hitlist, misslist

    def updateResidueTypes(self):
        """
            Find the type of residue as notated in the Amino Acid definition
        """
        self.write("Updating Residue Types... ")
        for chain in self.protein.getChains():
            for residue in chain.get("residues"):
                name = residue.get("name")
                if name in AAS:
                    residue.set("type", 1)
                elif name == "WAT":
                    residue.set("type", 3)
                elif name in NAS:
                    residue.set("type", 4)
                else: # Residue is a ligand or unknown
                    residue.set("type", 2)

        self.write("Done\n")

    def updateSSbridges(self):
        """
            Check for SS-bridge partners, and if present, set appropriate
            partners
        """
        self.write("Updating SS bridges...\n")
        SGpartners = {}
        for residue in self.protein.getResidues():
            if isinstance(residue, CYS):
                atom = residue.getAtom("SG")
                if atom != None:
                    SGpartners[atom] = []

        for atom in SGpartners:
            for partner in SGpartners:
                if atom == partner or SGpartners[atom] != []: continue
                dist = distance(atom.getCoords(), partner.getCoords())
                if dist < BONDED_SS_LIMIT:
                    SGpartners[atom].append(partner)
                    SGpartners[partner].append(atom)

        for atom in SGpartners:
            res1 = atom.get("residue")
            numpartners = len(SGpartners[atom])
            if numpartners == 1:
                partner = SGpartners[atom][0]
                res2 = partner.get("residue")
                res1.set("SSbonded", 1)
                res1.set("SSbondedpartner", partner)
                self.applyPatch("CYX", res1)
                self.write("%s - %s\n" % (res1, res2), 1)
            elif numpartners > 1:
                error = "WARNING: %s has multiple potential " % res1
                error += "SS-bridge partners\n"
                self.write(error, 1)
                self.warnings.append(error)
            elif numpartners == 0:
                self.write("%s is a free cysteine\n" % res1, 1)
        self.write("Done.\n")

    def updateInternalBonds(self):
        """
            Update the internal bonding network using the reference
            objects in each atom.
        """
        for residue in self.protein.getResidues():
            if isinstance(residue, (Amino, WAT, Nucleic)):
                for atom in residue.getAtoms():
                    if not atom.hasReference(): continue
                    for bond in atom.reference.bonds:
                        if not residue.hasAtom(bond): continue
                        bondatom = residue.getAtom(bond)
                        if bondatom not in atom.bonds:
                            atom.addBond(bondatom)

    def updateBonds(self):
        """
            Update the bonding network of the protein.  This happens
            in 3 steps:
              1.  Applying the PEPTIDE patch to all Amino residues
                  so as to add reference for the N(i+1) and C(i-1)
                  atoms
              2.  UpdateInternalBonds for inter-residue linking
              3.  Set the links to the N(i+1) and C(i-1) atoms
        """

        # Apply the peptide patch

        for residue in self.protein.getResidues():
            if isinstance(residue, Amino):
                if residue.isNterm or residue.isCterm:
                    continue
                else:
                    self.applyPatch("PEPTIDE", residue)

        # Update all internal bonds

        self.updateInternalBonds()

        # Set the peptide bond pointers

        for chain in self.protein.getChains():
            for i in range(chain.numResidues() - 1):
                res1 = chain.residues[i]
                res2 = chain.residues[i + 1]
                if not isinstance(res1, Amino) or not isinstance(res2, Amino):
                    continue
                atom1 = res1.getAtom("C")
                atom2 = res2.getAtom("N")

                if atom1 != None:
                    res2.peptideC = atom1
                if atom2 != None:
                    res1.peptideN = atom2
                if atom1 == None or atom2 == None:
                    continue

                if distance(atom1.getCoords(), atom2.getCoords()) > PEPTIDE_DIST:
                    text = "Gap in backbone detected between %s and %s!\n" % \
                           (res1, res2)
                    self.write(text, 1)
                    self.warnings.append(text)
                    res2.peptideC = None
                    res1.peptideN = None

    def applyPatch(self, patchname, residue):
        """
            Apply a patch to the given residue.  This is one of the key
            functions in PDB2PQR.  A similar function appears in
            definitions.py - that version is needed for residue level
            subtitutions so certain protonation states (i.e. CYM, HSE)
            are detectatble on input.

            This version looks up the particular patch name in the
            patchmap stored in the protein, and then applies the
            various commands to the reference and actual residue
            structures.

            See the inline comments for a more detailed explanation.

            Parameters
                patchname:  The name of the patch (string)
                residue:    The residue to apply the patch to (residue)
        """
        if patchname not in self.protein.patchmap:
            raise PDBInternalError("Unable to find patch %s!" % patchname)

        self.write('PATCH INFO: %s patched with %s\n' % (residue,patchname),1)

        # Make a copy of the reference, i.e. a new reference for
        # this patch.  Two examples:
        #     PEPTIDE is a special case, as it applies to
        #             every residue.
        #     CTERM only applies to one specific residue, so a
        #             deep copy is used.

        if patchname == "PEPTIDE":
            newreference = residue.reference
        else:
            newreference = copy.deepcopy(residue.reference)

        patch = self.protein.patchmap[patchname]

        # Add atoms from patch
        for atomname in patch.map:
            newreference.map[atomname] = patch.map[atomname]
            for bond in patch.map[atomname].bonds:
                if bond not in newreference.map: continue
                if atomname not in newreference.map[bond].bonds:
                    newreference.map[bond].bonds.append(atomname)

        # Remove atoms as directed by patch

        for remove in patch.remove:
            if remove in residue.map: residue.removeAtom(remove)
            if remove not in newreference.map: continue
            removebonds = newreference.map[remove].bonds
            del newreference.map[remove]
            for bond in removebonds:
                index = newreference.map[bond].bonds.index(remove)
                del newreference.map[bond].bonds[index]

        # Add the new dihedrals

        for dihedral in patch.dihedrals:
            newreference.dihedrals.append(dihedral)

        # Point at the new reference

        residue.reference = newreference
        residue.patches.append(patchname)

        # Rename atoms as directed by patch

        for atom in residue.getAtoms():
            if atom.name in patch.altnames:
                residue.renameAtom(atom.name, patch.altnames[atom.name])

        # Replace each atom's reference with the new one
        for atomname in residue.map:
            if newreference.hasAtom(atomname):
                atom = residue.getAtom(atomname)
                atom.reference = newreference.map[atomname]

    def setStates(self):
        """
            Set the state of each residue.  This is the last step
            before assigning the forcefield, but is necessary so
            as to distinguish between various protonation states.

            See aa.py for residue-specific functions.
        """
        for residue in self.protein.getResidues():
            if isinstance(residue, (Amino, Nucleic)):
                residue.setState()

    def assignTermini(self, chain, neutraln=False, neutralc=False):
        """
            Assign the termini for the given chain by looking at
            the start and end residues.
        """

        if len(chain.residues) == 0:
            text = "Error: chain \"%s\" has 0 residues!" % chain.chainID
            raise PDBInputError(text)

        # Set the N-Terminus/ 5' Terminus

        res0 = chain.residues[0]
        if isinstance(res0, Amino):
            res0.set("isNterm", 1)
            if isinstance(res0, PRO):
                self.applyPatch("NEUTRAL-NTERM", res0) # MODIFIED
            elif neutraln:
                self.applyPatch("NEUTRAL-NTERM", res0)
            else:
                self.applyPatch("NTERM", res0)
        elif isinstance(res0, Nucleic):
            res0.set("is5term", 1)
            self.applyPatch("5TERM", res0)

        # Set the C-Terminus/ 3' Terminus

        reslast = chain.residues[-1]
        if isinstance(reslast, Amino):
            reslast.set("isCterm", 1)
            if neutralc:
                self.applyPatch("NEUTRAL-CTERM", reslast)
            else:
                self.applyPatch("CTERM", reslast)
        elif isinstance(reslast, Nucleic):
            reslast.set("is3term", 1)
            self.applyPatch("3TERM", reslast)
        else:
            for i in range(len(chain.residues)):
                resthis = chain.residues[-1 - i]
                if isinstance(resthis, Amino):
                    resthis.set("isCterm", 1)
                    if neutralc:
                        self.applyPatch("NEUTRAL-CTERM", resthis)
                    else:
                        self.applyPatch("CTERM", resthis)
                    break
                elif resthis.name in ["NH2", "NME"]: break
                elif isinstance(resthis, Nucleic):
                    resthis.set("is3term", 1)
                    self.applyPatch("3TERM", resthis)
                    break

    def setTermini(self, neutraln=False, neutralc=False):
        """
            Set the termini for the protein. First set all known
            termini by looking at the ends of the chain. Then
            examine each residue, looking for internal chain breaks.
        """

        self.write("Setting the termini... \n")

        # First assign the known termini

        for chain in self.protein.getChains():
            self.assignTermini(chain, neutraln, neutralc)

        # Now determine if there are any hidden chains

        letters = string.ascii_uppercase + string.ascii_lowercase
        c = 0

        while c < len(self.protein.getChains()):
            chain = self.protein.chains[c]
            reslist = []

            origlist = []

            # origlist holds the original residue list for the chain

            for residue in chain.getResidues():
                origlist.append(residue)

            for residue in origlist:
                reslist.append(residue)
                oldid = residue.chainID

                # Look for ending termini

                fixflag = 0
                if isinstance(residue, Amino):
                    if (residue.hasAtom("OXT") and not residue.isCterm):
                        fixflag = 1

                elif isinstance(residue, Nucleic):
                    if ((residue.hasAtom("H3T") or residue.name.endswith("3"))\
                      and not residue.is3term):
                        fixflag = 1

                if fixflag:

                    # Get an available chain ID
                    chainid = letters[0]
                    id = 0
                    idLength = 1
                    while chainid in self.protein.chainmap:
                        id += 1
                        if id >= len(letters):
                            idLength += 1
                            id = 0
                        chainid = letters[id] * idLength

                    if(idLength > 1):
                        message = 'Warning: Reusing chain id: ' + chainid[0] + '\n'
                        self.write(message)

                    # Make a new chain with these residues
                    newchain = Chain(chainid[0])

                    self.protein.chainmap[chainid] = newchain
                    self.protein.chains.insert(c, newchain)

                    for res in reslist:
                        newchain.addResidue(res)
                        chain.residues.remove(res)
                        res.setChainID(chainid[0])

                    self.assignTermini(chain, neutraln, neutralc)
                    self.assignTermini(newchain, neutraln, neutralc)

                    reslist = []
                    c += 1

            c += 1

        # Update the final chain's chainID if it is "" unless it's all water

        if "" in self.protein.chainmap:

            notwat = 0
            for res in chain.residues:
                if not isinstance(res, WAT):
                    notwat = 1
                    break

            if notwat == 0:
                self.write("Done.\n")
                return

            chain = self.protein.chainmap[""]
            chainid = letters[0]
            id = 0
            idLength = 1
            while chainid in self.protein.chainmap:
                id += 1
                if id >= len(letters):
                    idLength += 1
                    id = 0
                chainid = letters[id] * idLength

            if(idLength > 1):
                message = 'Warning: Reusing chain id: ' + chainid[0] + '\n'
                self.write(message)

            # Use the new chainID

            self.protein.chainmap[chainid] = chain
            del self.protein.chainmap[""]

            for res in chain.residues:
                res.setChainID(chainid[0])

        self.write("Done.\n")


    def findMissingHeavy(self):
        """
            Repair residues that contain missing heavy (non-Hydrogen) atoms
        """
        self.write("Checking for missing heavy atoms... \n")
        misscount = 0
        heavycount = 0
        for residue in self.protein.getResidues():
            if not isinstance(residue, (Amino, Nucleic)): continue

            # Check for Missing Heavy Atoms

            for refatomname in residue.reference.map:
                if refatomname.startswith("H"): continue
                if refatomname in ["N+1", "C-1"]: continue
                if refatomname in ["O1P", "O2P"]:
                    if residue.hasAtom("OP1") and residue.hasAtom("OP2"): continue
                heavycount += 1
                if not residue.hasAtom(refatomname):
                    self.write("Missing %s in %s\n" % \
                               (refatomname, residue), 1)
                    misscount += 1
                    residue.addMissing(refatomname)

            # Check for Extra Atoms

            atomlist = []
            for atom in residue.get("atoms"):
                atomlist.append(atom)

            for atom in atomlist:
                atomname = atom.get("name")
                if atomname in ["OP1", "OP2"] and residue.reference.hasAtom("O1P") \
                    and residue.reference.hasAtom("O2P"): continue
                if not residue.reference.hasAtom(atomname):
                    self.write("Extra atom %s in %s! - " % \
                               (atomname, residue), 1)
                    residue.removeAtom(atomname)
                    self.write("Deleted this atom.\n")

        if heavycount == 0:
            raise PDBInputError("No heavy atoms found. " +
                                "You may also see this message if PDB2PQR does not have parameters for any residue in your protein.")

        misspct = 100.0 * float(misscount) / heavycount
        if misspct > REPAIR_LIMIT:
            error = "This PDB file is missing too many (%i out of " % misscount
            error += "%i, %.2f%%) heavy atoms to accurately repair the file.  " % \
                     (heavycount, misspct)
            error += "The current repair limit is set at %i%%. " % REPAIR_LIMIT
            error += "You may also see this message if PDB2PQR does not have parameters for enough residues in your protein."
            raise PDBInputError(error)
        elif misscount > 0:
            self.write("Missing %i out of %i heavy atoms (%.2f percent) - " % \
                       (misscount, heavycount, misspct))
            self.write("Will attempt to repair.\n")
            self.repairHeavy()
        else:
            self.write("No heavy atoms found missing - Done.\n")

    @staticmethod
    def rebuildTetrahedral(residue, atomname):
        """
            Rebuild a tetrahedral hydrogen group.  This is necessary
            due to the shortcomings of the quatfit routine - given a
            tetrahedral geometry and two existing hydrogens, the
            quatfit routines have two potential solutions.  This function
            uses basic tetrahedral geometry to fix this issue.

            Parameters
                residue:  The residue in question (residue)
                atomname: The atomname to add (string)
            Returns
                1 if successful, 0 otherwise
        """

        hcount = 0
        nextatomname = None

        atomref = residue.reference.map.get(atomname)
        if atomref is None:
            return False
        bondname = atomref.bonds[0]

        # Return if the bonded atom does not exist

        if not residue.hasAtom(bondname):
            return False

        # This group is tetrahedral if bondatom has 4 bonds,
        #  3 of which are hydrogens

        for bond in residue.reference.map[bondname].bonds:
            if bond.startswith("H"):
                hcount += 1
            elif bond != 'C-1' and bond != 'N+1':
                nextatomname = bond

        # Check if this is a tetrahedral group

        if hcount != 3 or nextatomname == None:
            return False

        # Now rebuild according to the tetrahedral geometry

        bondatom = residue.getAtom(bondname)
        nextatom = residue.getAtom(nextatomname)
        numbonds = len(bondatom.bonds)

        if numbonds == 1:

            # Place according to two atoms

            coords = [bondatom.getCoords(), nextatom.getCoords()]
            refcoords = [residue.reference.map[bondname].getCoords(), \
                         residue.reference.map[nextatomname].getCoords()]
            refatomcoords = atomref.getCoords()
            newcoords = findCoordinates(2, coords, refcoords, refatomcoords)
            residue.createAtom(atomname, newcoords)

            # For LEU and ILE residues only: make sure the Hydrogens are in staggered conformation instead of eclipsed.

            if isinstance(residue, LEU):
                hcoords = newcoords
                cbatom = residue.getAtom('CB')
                ang = getDihedral(cbatom.getCoords(), nextatom.getCoords(), bondatom.getCoords(), hcoords)
                diffangle = 60 - ang
                residue.rotateTetrahedral(nextatom, bondatom, diffangle)

            elif isinstance(residue, ILE):
                hcoords = newcoords
                cg1atom = residue.getAtom('CG1')
                cbatom = residue.getAtom('CB')
                if bondatom.name == 'CD1':
                    ang = getDihedral(cbatom.getCoords(), nextatom.getCoords(), bondatom.getCoords(), hcoords)
                elif bondatom.name == 'CG2':
                    ang = getDihedral(cg1atom.getCoords(), nextatom.getCoords(), bondatom.getCoords(), hcoords)
                else:
                    ang = getDihedral(cbatom.getCoords(), nextatom.getCoords(), bondatom.getCoords(), hcoords)

                diffangle = 60 - ang
                residue.rotateTetrahedral(nextatom, bondatom, diffangle)

            return 1

        elif numbonds == 2:

            # Get the single hydrogen coordinates

            hatom = None
            for bond in bondatom.reference.bonds:
                if residue.hasAtom(bond) and bond.startswith("H"):
                    hatom = residue.getAtom(bond)
                    break

            # Use the existing hydrogen and rotate about the bond

            residue.rotateTetrahedral(nextatom, bondatom, 120)
            newcoords = hatom.getCoords()
            residue.rotateTetrahedral(nextatom, bondatom, -120)
            residue.createAtom(atomname, newcoords)

            return 1

        elif numbonds == 3:

            # Find the one spot the atom can be

            hatoms = []
            for bond in bondatom.reference.bonds:
                if residue.hasAtom(bond) and bond.startswith("H"):
                    hatoms.append(residue.getAtom(bond))

            # If this is more than two something is wrong

            if len(hatoms) != 2: return 0

            # Use the existing hydrogen and rotate about the bond

            residue.rotateTetrahedral(nextatom, bondatom, 120)
            newcoords1 = hatoms[0].getCoords()
            residue.rotateTetrahedral(nextatom, bondatom, 120)
            newcoords2 = hatoms[0].getCoords()
            residue.rotateTetrahedral(nextatom, bondatom, 120)

            # Determine which one hatoms[1] is not in

            if distance(hatoms[1].getCoords(), newcoords1) > 0.1:
                residue.createAtom(atomname, newcoords1)
            else:
                residue.createAtom(atomname, newcoords2)

            return 1

    def addHydrogens(self):
        """
            Add the hydrogens to the protein.  This requires either
            the rebuildTetrahedral function for tetrahedral geometries
            or the standard quatfit methods.  These methods use three
            nearby bonds to rebuild the atom; the closer the bonds, the
            more accurate the results.  As such the peptide bonds are
            used when available.
        """
        count = 0
        self.write("Adding hydrogens to the protein...\n")
        for residue in self.protein.getResidues():
            if not isinstance(residue, (Amino, Nucleic)):
                continue
            for atomname in residue.reference.map:
                if not atomname.startswith("H"):
                    continue
                if residue.hasAtom(atomname):
                    continue
                if isinstance(residue, CYS) and residue.SSbonded and atomname == "HG":
                    continue

                # If this hydrogen is part of a tetrahedral group,
                #  follow a different codepath

                if Routines.rebuildTetrahedral(residue, atomname):
                    count += 1
                    continue

                # Otherwise use the standard quatfit methods

                coords = []
                refcoords = []

                refatomcoords = residue.reference.map[atomname].getCoords()
                bondlist = residue.reference.getNearestBonds(atomname)

                for bond in bondlist:
                    if bond == "N+1":
                        atom = residue.peptideN
                    elif bond == "C-1":
                        atom = residue.peptideC
                    else:
                        atom = residue.getAtom(bond)

                    if atom == None:
                        continue

                    # Get coordinates, reference coordinates

                    coords.append(atom.getCoords())
                    refcoords.append(residue.reference.map[bond].getCoords())

                    # Exit if we have enough atoms

                    if len(coords) == 3:
                        break

                if len(coords) == 3:
                    newcoords = findCoordinates(3, coords, refcoords, refatomcoords)
                    residue.createAtom(atomname, newcoords)
                    count += 1
                else:
                    self.write("Couldn't rebuild %s in %s!\n" % (atomname, residue), 1)            

        self.write(" Added %i hydrogen atoms.\n" % count)
        

    def removeHydrogens(self):
        self.write("Stripping hydrogens from the protein...\n")

        for residue in self.protein.getResidues():
            if not isinstance(residue, (Amino, Nucleic)):
                continue
            for atom in residue.atoms[:]:
                if atom.isHydrogen():
                    residue.removeAtom(atom.name)

    def repairHeavy(self):
        """
            Repair all heavy atoms.  Unfortunately the first time we
            get to an atom we might not be able to rebuild it - it
            might depend on other atoms to be rebuild first (think side
            chains).  As such a 'seenmap' is used to keep track of what
            we've already seen and subsequent attempts to rebuild the
            atom.
        """
        self.write("Rebuilding missing heavy atoms... \n")
        for residue in self.protein.getResidues():
            if not isinstance(residue, (Amino, Nucleic)):
                continue
            missing = residue.get("missing")
            if missing == []: continue

            # Initialize some variables

            seenmap = {}
            nummissing = len(missing)

            while len(missing) > 0:
                coords = []
                refcoords = []

                atomname = missing.pop(0)
                refatomcoords = residue.reference.map[atomname].getCoords()
                bondlist = residue.reference.getNearestBonds(atomname)

                for bond in bondlist:
                    if bond == "N+1": atom = residue.peptideN
                    elif bond == "C-1": atom = residue.peptideC
                    else: atom = residue.getAtom(bond)

                    if atom == None: continue

                    # Get coordinates, reference coordinates

                    coords.append(atom.getCoords())
                    refcoords.append(residue.reference.map[bond].getCoords())

                    # Exit if we have enough atoms

                    if len(coords) == 3: break

                # We might need other atoms to be rebuilt first

                if len(coords) < 3:
                    try:
                        seenmap[atomname] += 1
                    except KeyError:
                        seenmap[atomname] = 1

                    missing.append(atomname)
                    if seenmap[atomname] > nummissing:
                        text = "Too few atoms present to reconstruct or cap residue %s in structure!\n" % (residue)
                        text += "This error is generally caused by missing backbone atoms in this protein;\n"
                        text += "you must use an external program to complete gaps in the protein backbone.\n"
                        text += "Heavy atoms missing from %s: " % (residue)
                        text += ' '.join(missing)
                        raise PDBInputError(text)

                else: # Rebuild the atom
                    newcoords = findCoordinates(3, coords, refcoords, refatomcoords)
                    residue.createAtom(atomname, newcoords)
                    self.write("Added %s to %s at coordinates" % (atomname, residue), 1)
                    self.write(" %.3f %.3f %.3f\n" % \
                           (newcoords[0], newcoords[1], newcoords[2]))
                    info = "#1.1 Warning: Added {0} to {1}".format(atomname, residue)
                    with open('LOGFILE', 'a') as logfile:
                        logfile.write(info + '\n')


        self.write("Done.\n")

    def setReferenceDistance(self):
        """
            Set the distance to the CA atom in the residue.
            This is necessary for determining which atoms are
            allowed to move during rotations.  Uses the
            shortestPath algorithm found in utilities.py.
        """
        for residue in self.protein.getResidues():
            if not isinstance(residue, Amino): continue

            # Initialize some variables

            map = {}
            caatom = residue.getAtom("CA")

            if caatom == None:
                text = "Cannot set references to %s without CA atom!\n"
                raise PDBInputError(text)

            # Set up the linked map

            for atom in residue.getAtoms():
                map[atom] = atom.bonds

            # Run the algorithm

            for atom in residue.getAtoms():
                if atom.isBackbone():
                    atom.refdistance = -1
                elif residue.isCterm and atom.name == "HO":   # special case for HO in Cterm
                    atom.refdistance = 3
                elif residue.isNterm and (atom.name == "H3" or atom.name == "H2"):  # special case for H2 or H3 in Nterm
                    atom.refdistance = 2
                else:
                    atom.refdistance = len(shortestPath(map, atom, caatom)) - 1

    def getbumpscore(self, residue):
        """Get an bump score for the current structure"""

        # Do some setup

        self.cells = Cells(CELL_SIZE)
        self.cells.assignCells(self.protein)

        self.calculateDihedralAngles()
        self.setDonorsAndAcceptors()
        self.updateInternalBonds()
        self.setReferenceDistance()
        bumpscore = 0.0
        #for residue in self.protein.getResidues():
        if not isinstance(residue, Amino): return 0.0
        # Initialize variables

        conflictnames = []

        for atom in residue.getAtoms():
            atomname = atom.name
            #if not atom.added: continue
            if atomname[0] != "H": continue
            #if atom.optimizeable: continue
            #print atomname,atom.optimizeable,atom.added
            bumpscore = bumpscore + self.getbumpscore_atom(atom)
        return bumpscore


    def getbumpscore_atom(self, atom):
        """
            Find nearby atoms for conflict-checking.  Uses
            neighboring cells to compare atoms rather than an all
            versus all O(n^2) algorithm, which saves a great deal
            of time.  There are several instances where we ignore
            potential conflicts; these include donor/acceptor pairs,
            atoms in the same residue, and bonded CYS bridges.

            Parameters
                atom:  Find nearby atoms to this atom (Atom)
            Returns
                bumpscore: a bump score sum((dist-cutoff)**20 for all near atoms

            Jens rewrote this function from findNearbyAtoms to
            be usable for detecting bumps for optimzable hydrogens
        """
        # Initialize some variables
        nearatoms = {}
        residue = atom.residue
        #print 'Cutoff distance',cutoff
        atom_size = BUMP_HYDROGEN_SIZE if atom.isHydrogen() else BUMP_HEAVY_SIZE

        # Get atoms from nearby cells

        closeatoms = self.cells.getNearCells(atom)

        # Loop through and see if any are within the cutoff

        bumpscore = 0.0
        for closeatom in closeatoms:
            closeresidue = closeatom.residue
            if closeresidue == residue and (closeatom in atom.bonds or atom in closeatom.bonds):
                continue

            if not isinstance(closeresidue, Amino):
                continue
            if isinstance(residue, CYS):
                if residue.SSbondedpartner == closeatom: continue

            # Also ignore if this is a donor/acceptor pair
            pair_ignored = False
            if atom.isHydrogen() and len(atom.bonds) != 0 and atom.bonds[0].hdonor \
               and closeatom.hacceptor:
                #pair_ignored = True
                continue
            if closeatom.isHydrogen() and len(closeatom.bonds) != 0 and closeatom.bonds[0].hdonor \
                   and atom.hacceptor:
                #pair_ignored = True
                continue

            dist = distance(atom.getCoords(), closeatom.getCoords())
            other_size = BUMP_HYDROGEN_SIZE if closeatom.isHydrogen() else BUMP_HEAVY_SIZE
            cutoff = atom_size + other_size
            if dist < cutoff:
                bumpscore = bumpscore + 1000.0
                if pair_ignored:
                    self.write('This bump is a donor/acceptor pair.\n')
#                if heavy_not_ignored:
#                    print 'Bumped {0} against {1} within residue'.format(atom.name, closeatom.name)
                #nearatoms[closeatom] = (dist-cutoff)**2
        self.write('BUMPSCORE ' + str(bumpscore) + '\n')
        return bumpscore


    def debumpProtein(self):
        """
            Make sure that none of the added atoms were rebuilt
            on top of existing atoms.  See each called function
            for more information.
        """

        self.write("Checking if we must debump any residues... \n")

        # Do some setup

        self.cells = Cells(CELL_SIZE)
        self.cells.assignCells(self.protein)

        self.calculateDihedralAngles()
        self.setDonorsAndAcceptors()
        self.updateInternalBonds()
        self.setReferenceDistance()

        # Determine which residues to debump

        for residue in self.protein.getResidues():
            if not isinstance(residue, Amino): continue

            # Initialize variables

            conflictnames = self.findResidueConflicts(residue, True)

            if not conflictnames:
                continue

            # Otherwise debump the residue

            self.write("Starting to debump %s...\n" % residue, 1)
            self.write("Debumping cutoffs: %2.1f for heavy-heavy, %2.1f for hydrogen-heavy, and %2.1f for hydrogen-hydrogen.\n" %
                       (BUMP_HEAVY_SIZE*2,
                        BUMP_HYDROGEN_SIZE+BUMP_HEAVY_SIZE,
                        BUMP_HYDROGEN_SIZE*2), 1)
            if self.debumpResidue(residue, conflictnames):
                self.write("Debumping Successful!\n\n", 1)
            else:
                text = "WARNING: Unable to debump %s\n" % residue
                self.write("********\n%s********\n\n" % text)
                self.warnings.append(text)

        self.write("Done.\n")

    def findResidueConflicts(self, residue, writeConflictInfo=False):
        conflictnames = []
        for atom in residue.getAtoms():
            atomname = atom.name
            if not atom.added: continue
            if atomname == "H": continue
            if atom.optimizeable: continue

            nearatoms = self.findNearbyAtoms(atom)

            # If something is too close, we must debump the residue

            if nearatoms != {}:
                conflictnames.append(atomname)
                if writeConflictInfo:
                    for repatom in nearatoms:
                        self.write("%s %s is too close to %s %s\n" % \
                                   (residue, atomname, repatom.residue, repatom.name), 1)

        return conflictnames

    def scoreDihedralAngle(self, residue, anglenum):
        score = 0
        atomnames = residue.reference.dihedrals[anglenum].split()
        pivot = atomnames[2]
        moveablenames = self.getMoveableNames(residue, pivot)
        for name in moveablenames:
            nearatoms = self.findNearbyAtoms(residue.getAtom(name))
            for v in nearatoms.values():
                score += v

        return score


    def debumpResidue(self, residue, conflictnames):
        """
            Debump a specific residue.  Only should be called
            if the residue has been detected to have a conflict.
            If called, try to rotate about dihedral angles to
            resolve the conflict.

            Parameters
                residue:  The residue in question
                conflictnames:  A list of atomnames that were
                                rebuilt too close to other atoms
            Returns
                True if successful, False otherwise
        """

        # Initialize some variables

        ANGLE_STEPS = 72
        ANGLE_STEP_SIZE = float(360 // ANGLE_STEPS)

        ANGLE_TEST_COUNT = 10

        anglenum = -1
        currentConflictNames = conflictnames

        EPSILON = 0.0000001

        # Try (up to 10 times) to find a workable solution

        for _ in range(ANGLE_TEST_COUNT):

            anglenum = self.pickDihedralAngle(residue, currentConflictNames, anglenum)

            if anglenum == -1: return False

            self.write("Using dihedral angle number %i to debump the residue.\n" % anglenum, 1)

            bestscore = self.scoreDihedralAngle(residue, anglenum)
            foundImprovement = False
            bestangle = originalAngle = residue.dihedrals[anglenum]

            #Skip the first angle as it's already known.
            for i in xrange(1, ANGLE_STEPS):
                newangle = originalAngle + (ANGLE_STEP_SIZE * i)
                self.setDihedralAngle(residue, anglenum, newangle)

                # Check for conflicts

                score = self.scoreDihedralAngle(residue, anglenum)

                if score == 0:
                    if not self.findResidueConflicts(residue):
                        self.write("No conflicts found at angle "+repr(newangle)+"\n", 1)
                        return True
                    else:
                        bestangle = newangle
                        foundImprovement = True
                        break

                # Set the best angle
                elif score < bestscore:
                    diff = abs(bestscore - score)
                    #Don't update if it's effectively a tie
                    if diff > EPSILON:
                        bestscore = score
                        bestangle = newangle
                        foundImprovement = True

            self.setDihedralAngle(residue, anglenum, bestangle)
            currentConflictNames = self.findResidueConflicts(residue)

            if foundImprovement:
                self.write("Best score of "+repr(bestscore)+" at angle "+repr(bestangle)+". New conflict set: ", 1)
                self.write(str(currentConflictNames)+"\n", 1)
            else:
                self.write("No improvement found for this dihedral angle.\n", 1)

        # If we're here, debumping was unsuccessful

        return False

    def calculateDihedralAngles(self):
        """
            Calculate the dihedral angle for every residue within the protein
        """
        for residue in self.protein.getResidues():
            if not isinstance(residue, Amino): continue
            residue.dihedrals = []

            refangles = residue.reference.dihedrals
            for di in refangles:
                coords = []
                atoms = di.split()
                for i in range(4):
                    atomname = atoms[i]
                    if residue.hasAtom(atomname):
                        coords.append(residue.getAtom(atomname).getCoords())

                if len(coords) == 4: angle = getDihedral(coords[0], coords[1], coords[2], coords[3])
                else: angle = None

                residue.addDihedralAngle(angle)

    def getClosestAtom(self, atom):
        """
            Get the closest atom that does not form a donor/acceptor pair.
            Used to detect potential conflicts.

            NOTE:  Cells must be set before using this function.

            Parameters
                atom:  The atom in question (Atom)
            Returns
                bestatom:  The closest atom to the input atom that does not
                           satisfy a donor/acceptor pair.
        """
        # Initialize some variables

        bestdist = 999.99
        bestwatdist = 999.99
        bestatom = None
        bestwatatom = None
        residue = atom.residue

        # Get atoms from nearby cells

        closeatoms = self.cells.getNearCells(atom)

        # Loop through and see which is the closest

        for closeatom in closeatoms:
            closeresidue = closeatom.residue
            if closeresidue == residue: continue
            if not isinstance(closeresidue, (Amino,WAT)): continue
            if isinstance(residue, CYS):
                if residue.SSbondedpartner == closeatom: continue

            # Also ignore if this is a donor/acceptor pair

            if atom.isHydrogen() and atom.bonds[0].hdonor \
               and closeatom.hacceptor: continue
            if closeatom.isHydrogen() and closeatom.bonds[0].hdonor \
                   and atom.hacceptor:
                continue

            dist = distance(atom.getCoords(), closeatom.getCoords())

            if isinstance(closeresidue, WAT):
                if dist < bestwatdist:
                    bestwatdist = dist
                    bestwatatom = closeatom
            else:
                if dist < bestdist:
                    bestdist = dist
                    bestatom = closeatom

        if bestdist > bestwatdist:
            txt = "Warning: %s in %s skipped when optimizing %s in %s\n" % (bestwatatom.name,
                                                                           bestwatatom.residue,
                                                                           atom.name, residue)
            if txt not in self.warnings:
                self.warnings.append(txt)

        return bestatom

    def findNearbyAtoms(self, atom):
        """
            Find nearby atoms for conflict-checking.  Uses
            neighboring cells to compare atoms rather than an all
            versus all O(n^2) algorithm, which saves a great deal
            of time.  There are several instances where we ignore
            potential conflicts; these include donor/acceptor pairs,
            atoms in the same residue, and bonded CYS bridges.

            Parameters
                atom:  Find nearby atoms to this atom (Atom)
            Returns
                nearatoms:  A dictionary of <Atom too close> to <amount of overlap for that atom>.
        """
        # Initialize some variables

        nearatoms = {}
        residue = atom.residue
        atom_size = BUMP_HYDROGEN_SIZE if atom.isHydrogen() else BUMP_HEAVY_SIZE

        # Get atoms from nearby cells

        closeatoms = self.cells.getNearCells(atom)

        # Loop through and see if any are within the cutoff

        for closeatom in closeatoms:
            closeresidue = closeatom.residue
            if closeresidue == residue and (closeatom in atom.bonds or atom in closeatom.bonds):
                continue

            if not isinstance(closeresidue, (Amino, WAT)):
                continue
            if isinstance(residue, CYS) and residue.SSbondedpartner == closeatom:
                continue

            # Also ignore if this is a donor/acceptor pair

            if (atom.isHydrogen() and len(atom.bonds) != 0 and
                atom.bonds[0].hdonor and closeatom.hacceptor):
                continue

            if (closeatom.isHydrogen() and len(closeatom.bonds) != 0 and
                closeatom.bonds[0].hdonor and atom.hacceptor):
                continue

            dist = distance(atom.getCoords(), closeatom.getCoords())
            other_size = BUMP_HYDROGEN_SIZE if closeatom.isHydrogen() else BUMP_HEAVY_SIZE
            cutoff = atom_size + other_size
            if dist < cutoff:
                nearatoms[closeatom] = cutoff - dist

        return nearatoms


    def pickDihedralAngle(self, residue, conflictnames, oldnum=None):
        """
            Choose an angle number to use in debumping

            Algorithm
                Instead of simply picking a random chiangle, this function
                uses a more intelligent method to improve efficiency.
                The algorithm uses the names of the conflicting atoms
                within the residue to determine which angle number
                has the best chance of fixing the problem(s). The method
                also insures that the same chiangle will not be run twice
                in a row.
            Parameters
                residue:    The residue that is being debumped (Residue)
                conflictnames: A list of atom names that are currently
                               conflicts (list)
                oldnum    : The old dihedral angle number (int)
            Returns
                bestnum    : The new dihedral angle number (int)
        """
        bestnum = -1
        best = 0

        iList = range(len(residue.dihedrals))
        #Make sure our testing is done round robin.
        if oldnum is not None and oldnum >= 0 and len(iList) > 0:
            del iList[oldnum]
            testDihedralIndecies = iList[oldnum:] + iList[:oldnum]
        else:
            testDihedralIndecies = iList

        for i in testDihedralIndecies:
            if i == oldnum: continue
            if residue.dihedrals[i] is None: continue

            score = 0
            atomnames = residue.reference.dihedrals[i].split()
            pivot = atomnames[2]

            moveablenames = self.getMoveableNames(residue, pivot)

            # If this pivot only moves the conflict atoms, pick it

            if conflictnames == moveablenames: return i

            # Otherwise find the pivot with the most matches

            for name in conflictnames:
                if name in moveablenames:
                    score += 1
                    if score > best:
                        best = score
                        bestnum = i

        # Return the best angle.  If none were found, return -1.

        return bestnum

    def setDihedralAngle(self, residue, anglenum, angle):
        """
            Rotate a residue about a given angle. Uses the quatfit
            methods to perform the matrix mathematics.

            Parameters
                residue:   The residue to rotate
                anglenum:  The number of the angle to rotate as
                           listed in residue.dihedrals
                angle:     The desired angle.

        """
        coordlist = []
        initcoords = []
        movecoords = []
        pivot = ""

        oldangle = residue.dihedrals[anglenum]
        diff = angle - oldangle

        atomnames = residue.reference.dihedrals[anglenum].split()

        pivot = atomnames[2]
        for atomname in atomnames:
            if residue.hasAtom(atomname):
                coordlist.append(residue.getAtom(atomname).getCoords())
            else:
                raise PDBInputError("Error occurred while trying to debump!")

        initcoords = subtract(coordlist[2], coordlist[1])

        moveablenames = self.getMoveableNames(residue, pivot)

        for name in moveablenames:
            atom = residue.getAtom(name)
            movecoords.append(subtract(atom.getCoords(), coordlist[1]))

        newcoords = qchichange(initcoords, movecoords, diff)

        for i in range(len(moveablenames)):
            atom = residue.getAtom(moveablenames[i])
            self.cells.removeCell(atom)
            x = (newcoords[i][0] + coordlist[1][0])
            y = (newcoords[i][1] + coordlist[1][1])
            z = (newcoords[i][2] + coordlist[1][2])
            atom.set("x", x)
            atom.set("y", y)
            atom.set("z", z)
            self.cells.addCell(atom)


        # Set the new angle

        coordlist = []
        for atomname in atomnames:
            if residue.hasAtom(atomname):
                coordlist.append(residue.getAtom(atomname).getCoords())
            else:
                raise PDBInputError("Error occurred while trying to debump!")

        di = getDihedral(coordlist[0], coordlist[1], coordlist[2], coordlist[3])
        residue.dihedrals[anglenum] = di

    def getMoveableNames(self, residue, pivot):
        """
            Return all atomnames that are further away than the
            pivot atom.

            Parameters
                residue:  The residue to use
                pivot:    The pivot atomname
        """
        movenames = []
        refdist = residue.getAtom(pivot).refdistance
        for atom in residue.getAtoms():
            if atom.refdistance > refdist:
                movenames.append(atom.name)
        return movenames

    def setDonorsAndAcceptors(self):
        """
            Set the donors and acceptors within the protein
        """
        for residue in self.protein.getResidues():
            residue.setDonorsAndAcceptors()

    def runPDB2PKA(self, ph, ff, pdblist, ligand, verbose, pdb2pka_params):
        if ff.lower() != 'parse':
            PDB2PKAError('PDB2PKA can only be run with the PARSE force field.')

        self.write("Running PDB2PKA and applying at pH %.2f... \n" % ph)

        import pka
        from pdb2pka import pka_routines
        init_params = pdb2pka_params.copy()
        init_params.pop('pairene')
        init_params.pop('clean_output')

        results = pka.pre_init(original_pdb_list=pdblist,
                               ff=ff,
                               verbose=verbose,
                               ligand=ligand,
                               **init_params)
        output_dir, protein, routines, forcefield,apbs_setup, ligand_titratable_groups, maps, sd = results

        mypkaRoutines = pka_routines.pKaRoutines(protein, routines, forcefield, apbs_setup, output_dir, maps, sd,
                                                 restart=pdb2pka_params.get('clean_output'),
                                                 pairene=pdb2pka_params.get('pairene'))

        print 'Doing full pKa calculation'
        mypkaRoutines.runpKa()

        pdb2pka_warnings = mypkaRoutines.warnings[:]

        self.warnings.extend(pdb2pka_warnings)

        residue_ph = {}
        for pka_residue_tuple, calc_ph in mypkaRoutines.ph_at_0_5.iteritems():
            tit_type, chain_id, number_str = pka_residue_tuple
            if tit_type == 'NTR':
                tit_type = 'N+'
            elif tit_type == 'CTR':
                tit_type = 'C-'

            key = ' '.join([tit_type, number_str, chain_id])
            residue_ph[key] = calc_ph

        pformat(residue_ph)

        self.apply_pka_values(ff, ph, residue_ph)

        self.write('Finished running PDB2PKA.\n')


    def runPROPKA(self, ph, ff, rootname, outname, options):
        """
            Run PROPKA on the current protein, setting protonation states to
            the correct values

            Parameters
               ph:  The desired pH of the system
               ff:  The forcefield name to be used
               outname: The name of the PQR outfile
        """
        self.write("Running PROPKA and applying at pH %.2f... \n" % ph)

        from propka30.Source.protein import Protein as pkaProtein
        from propka30.Source.pdb import readPDB as pkaReadPDB
        from propka30.Source.lib import residueList, setVerbose

        setVerbose(options.verbose)

        # Initialize some variables

        linelen = 70
        pkadic = {}

        # Reorder the atoms in each residue to start with N

        for residue in self.protein.getResidues():
            residue.reorder()

        # Make a string with all non-hydrogen atoms

        HFreeProteinFile = StringIO()

        for atom in self.protein.getAtoms():
            if not atom.isHydrogen():
                atomtxt = atom.getPDBString()
                atomtxt = atomtxt[:linelen]
                HFreeProteinFile.write(atomtxt)
                HFreeProteinFile.write('\n')


        HFreeProteinFile.seek(0)

        # Run PropKa

        atoms = pkaReadPDB('', file=HFreeProteinFile)

        # creating protein object
        myPkaProtein = pkaProtein(atoms=atoms, name=rootname, options=options)
        # calculating pKa values for ionizable residues
        myPkaProtein.calculatePKA(options=options)
        # printing pka file
        myPkaProtein.writePKA(options=options, filename=outname)

        # Parse the results
        # This is the method used to generate the summary in the first place.
        residue_list = residueList("propka1")
        for chain in myPkaProtein.chains:
            for residue_type in residue_list:
                for residue in chain.residues:
                    if residue.resName == residue_type:
                        #Strip out the extra space after C- or N+
                        key = string.strip('%s %s %s' % (string.strip(residue.resName),
                                                        residue.resNumb, residue.chainID))
                        pkadic[key] = residue.pKa_pro

        if len(pkadic) == 0:
            return

        # Now apply each pka to the appropriate residue
        self.apply_pka_values(ff, ph, pkadic)

        self.write("Done.\n")

    def apply_pka_values(self, ff, ph, pkadic):
        self.write('Applying pKa values at a pH of %.2f:\n' % ph)
        formatted_pkadict = pformat(pkadic)
        self.write(formatted_pkadict+'\n\n')

        warnings = []
        for residue in self.protein.getResidues():
            if not isinstance(residue, Amino):
                continue
            resname = residue.name
            resnum = residue.resSeq
            chainID = residue.chainID

            if residue.isNterm:
                key = "N+ %i %s" % (resnum, chainID)
                key = string.strip(key)
                if key in pkadic:
                    value = pkadic[key]
                    del pkadic[key]
                    if ph >= value:
                        if ff in ["amber", "charmm", "tyl06", "peoepb", "swanson"]:
                            warn = ("N-terminal %s" % key, "neutral")
                            warnings.append(warn)
                        else:
                            self.applyPatch("NEUTRAL-NTERM", residue)

            if residue.isCterm:
                key = "C- %i %s" % (resnum, chainID)
                key = string.strip(key)
                if key in pkadic:
                    value = pkadic[key]
                    del pkadic[key]
                    if ph < value:
                        if ff in ["amber", "charmm", "tyl06", "peoepb", "swanson"]:
                            warn = ("C-terminal %s" % key, "neutral")
                            warnings.append(warn)
                        else:
                            self.applyPatch("NEUTRAL-CTERM", residue)

            key = "%s %i %s" % (resname, resnum, chainID)
            key = string.strip(key)
            if key in pkadic:
                value = pkadic[key]
                del pkadic[key]
                if resname == "ARG" and ph >= value:
                    if ff == "parse":
                        self.applyPatch("AR0", residue)
                        txt = "WARNING: Neutral arginines are very rare. Please double\n"
                        self.warnings.append(txt)
                        self.write(txt)
                        txt = "         check your system and caculation setup.\n"
                        self.warnings.append(txt)
                        self.write(txt)
                    else:
                        warn = (key, "neutral")
                        warnings.append(warn)
                elif resname == "ASP" and ph < value:
                    if residue.isCterm and ff in ["amber", "tyl06", "swanson"]:
                        warn = (key, "Protonated at C-Terminal")
                        warnings.append(warn)
                    elif residue.isNterm and ff in ["amber", "tyl06", "swanson"]:
                        warn = (key, "Protonated at N-Terminal")
                        warnings.append(warn)
                    else:
                        self.applyPatch("ASH", residue)
                elif resname == "CYS" and ph >= value:
                    if ff == "charmm":
                        warn = (key, "negative")
                        warnings.append(warn)
                    else:
                        self.applyPatch("CYM", residue)
                elif resname == "GLU" and ph < value:
                    if residue.isCterm and ff in ["amber", "tyl06", "swanson"]:
                        warn = (key, "Protonated at C-Terminal")
                        warnings.append(warn)
                    elif residue.isNterm and ff in ["amber", "tyl06", "swanson"]:
                        warn = (key, "Protonated at N-Terminal")
                        warnings.append(warn)
                    else:
                        self.applyPatch("GLH", residue)
                elif resname == "HIS" and ph < value:
                    self.applyPatch("HIP", residue)
                elif resname == "LYS" and ph >= value:
                    if ff == "charmm":
                        warn = (key, "neutral")
                        warnings.append(warn)
                    elif ff in ["amber", "tyl06", "swanson"] and residue.get("isCterm"):
                        warn = (key, "neutral at C-Terminal")
                        warnings.append(warn)
                    elif ff == "tyl06" and residue.get("isNterm"):
                        warn = (key, "neutral at N-Terminal")
                        warnings.append(warn)
                    else:
                        self.applyPatch("LYN", residue)
                elif resname == "TYR" and ph >= value:
                    if ff in ["charmm", "amber", "tyl06", "peoepb", "swanson"]:
                        warn = (key, "negative")
                        warnings.append(warn)
                    else:
                        self.applyPatch("TYM", residue)

        if len(warnings) > 0:
            init = "WARNING: PDB2PKA determined the following residues to be\n"
            self.warnings.append(init)
            self.write(init)
            init = "         in a protonation state not supported by the\n"
            self.warnings.append(init)
            self.write(init)
            init = "         %s forcefield!\n" % ff
            self.warnings.append(init)
            self.write(init)
            init = "         All were reset to their standard pH 7.0 state.\n"
            self.warnings.append(init)
            self.write(init)
            self.warnings.append("\n")
            self.write('\n')
            for warn in warnings:
                text = "             %s (%s)\n" % (warn[0], warn[1])
                self.warnings.append(text)
                self.write(text)
            self.warnings.append("\n")
            self.write('\n')

        if len(pkadic) > 0:
            warn = "         PDB2PQR could not identify the following residues\n"
            self.warnings.append(warn)
            self.write(warn)
            warn = "         and residue numbers as returned by PROPKA or PDB2PKA:\n"
            self.warnings.append(warn)
            self.warnings.append("\n")
            self.write(warn)
            self.write('\n')
            for item in pkadic:
                text = "             %s\n" % item
                self.warnings.append(text)
                self.write(text)
            self.warnings.append("\n")
            self.write('\n')


class Cells:
    """
        The cells object provides a better way to search for nearby atoms. A
        pure all versus all search is O(n^2) - for every atom, every other atom
        must be searched.  This is rather inefficient, especially for large
        proteins where cells may be tens of angstroms apart.  The cell class
        breaks down the xyz protein space into several 3-D cells of desired
        size - then by simply examining atoms that fall into the adjacent
        cells one can quickly find nearby cells.

        NOTE:  Ideally this should be somehow separated from the routines
               object...
        """
    def __init__(self, cellsize):
        """
            Initialize the cells.

            Parameters
                cellsize:  The size of each cell (int)
        """
        self.cellmap = {}
        self.cellsize = cellsize

    def assignCells(self, protein):
        """
            Place each atom in a virtual cell for easy neighbor comparison
        """
        for atom in protein.getAtoms():
            atom.cell = None
            self.addCell(atom)

    def addCell(self, atom):
        """
            Add an atom to the cell

            Parameters
                atom:  The atom to add (atom)
        """
        size = self.cellsize

        x = atom.get("x")
        if x < 0:
            x = (int(x) - 1) / size * size
        else:
            x = int(x) / size * size

        y = atom.get("y")
        if y < 0:
            y = (int(y) - 1) / size * size
        else:
            y = int(y) / size * size

        z = atom.get("z")
        if z < 0:
            z = (int(z) - 1) / size * size
        else:
            z = int(z) / size * size

        key = (x, y, z)
        try:
            self.cellmap[key].append(atom)
        except KeyError:
            self.cellmap[key] = [atom]
        atom.set("cell", key)

    def removeCell(self, atom):
        """
             Remove the atom from a cell

             Parameters
                 atom:   The atom to add (atom)
        """
        oldcell = atom.get("cell")
        if oldcell == None: return
        atom.set("cell", None)
        self.cellmap[oldcell].remove(atom)

    def getNearCells(self, atom):
        """
            Find all atoms in bordering cells to an atom

            Parameters
                atom:  The atom to use (atom)
            Returns
                closeatoms:  A list of nearby atoms (list)
        """
        size = self.cellsize
        closeatoms = []
        cell = atom.get("cell")
        if cell == None:
            return closeatoms
        else:
            x = cell[0]
            y = cell[1]
            z = cell[2]
            for i in range(-1 * size, 2 * size, size):
                for j in range(-1 * size, 2 * size, size):
                    for k in range(-1 * size, 2 * size, size):
                        newkey = (x + i, y + j, z + k)
                        try:
                            newatoms = self.cellmap[newkey]
                            for atom2 in newatoms:
                                if atom == atom2: continue
                                closeatoms.append(atom2)
                        except KeyError: pass

            return closeatoms



"""
This was lifted straight from the scons documentation

http://www.scons.org/doc/HTML/scons-user.html#AEN2061

on September 5th, 2013

I've made a few modifications as needed.

Legal crap:
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import atexit


def bf_to_str(bf):
    """Convert an element of GetBuildFailures() to a string
    in a useful way."""
    import SCons.Errors

    if bf is None:  # unknown targets product None in list
        return "(unknown tgt)"
    elif isinstance(bf, SCons.Errors.StopError):
        return str(bf)
    elif bf.node:
        return str(bf.node) + ": " + bf.errstr
    elif bf.filename:
        return bf.filename + ": " + bf.errstr
    return "unknown failure: " + bf.errstr


def build_status():
    """Convert the build status to a 2-tuple, (status, msg)."""
    from SCons.Script import GetBuildFailures

    bf = GetBuildFailures()
    if bf:
        # bf is normally a list of build failures; if an element is None,
        # it's because of a target that scons doesn't know anything about.
        status = "failed"
        failures_message = "\n".join(
            ["Failed building %s" % bf_to_str(x) for x in bf if x is not None]
        )
    else:
        # if bf is None, the build completed successfully.
        status = "ok"
        failures_message = ""
    return (status, failures_message)


def display_build_status():
    """Display the build status.  Called by atexit.
    Here you could do all kinds of complicated things."""
    if GetOption("clean") or GetOption("help"):
        return

    status, failures_message = build_status()
    if status == "failed":
        print
        print "FAILED!!!!"  # could display alert, ring bell, etc.
    elif status == "ok":
        print
        print "Build succeeded."
    print failures_message


atexit.register(display_build_status)



#
# $Id: GromacsEM.py 290 2011-06-03 09:51:17Z nielsen $
#
# Python class for running Gromacs
#
import os


class GROMACS:
    def __init__(self, neededfiles, createdir=True):
        #
        # Find Gromacs
        #
        GMXBIN = None
        import os

        path = os.environ["PATH"]
        for dirname in path.split(":"):
            mdrun = os.path.join(dirname, "mdrun")
            if os.path.isfile(mdrun):
                GMXBIN = dirname
        if not GMXBIN:
            GMXBIN = "/usr/bin"
        environment = {"GMXBIN": GMXBIN}  # ,'GMXLIB':'/usr/share/gromacs/top/'}
        #
        # Set the environment
        #
        for key in environment.keys():
            os.putenv(key, environment[key])
        if createdir:
            #
            # Create the tempdir
            #
            import tempfile

            self.topdir = os.getcwd()
            tempfile.tempdir = self.topdir
            self.tmpdir = tempfile.mktemp()
            os.mkdir(self.tmpdir)
            print "Running in :", self.tmpdir
            os.chdir(self.tmpdir)
            #
            # Copy all the input files to the tmpdir
            #
            for file in neededfiles:
                os.system("cp " + file + " ./")
        #
        # Filenames
        #
        self.files = {}
        return

    #
    # ------
    #

    def cleanup(self):
        """Remove the tempdir"""
        import shutil

        os.chdir(self.topdir)
        shutil.rmtree(self.tmpdir)
        self.tmpdir = None
        return

    #
    # ---
    #

    def pdb2gmx(self, pdbfile, forcefield=0, ignore_Hs=True, auto_select_his=True):
        #
        # forcefield: 0 for Gromacs FF, 1 for Gromacs FF with all hyds.
        #
        print auto_select_his
        self.files["laststructure"] = "protein.gro"
        fd = open("pdb2gmx.scr", "w")
        if auto_select_his:
            if ignore_Hs:
                fd.write(
                    "pdb2gmx -f "
                    + pdbfile
                    + " -ignh -o "
                    + self.files["laststructure"]
                    + " << EOF\n"
                )
            else:
                fd.write(
                    "pdb2gmx -f "
                    + pdbfile
                    + " -o "
                    + self.files["laststructure"]
                    + " << EOF\n"
                )
        else:
            if ignore_Hs:
                fd.write(
                    "pdb2gmx -f "
                    + pdbfile
                    + " -ignh -his -o "
                    + self.files["laststructure"]
                    + " << EOF\n"
                )
            else:
                fd.write(
                    "pdb2gmx -f "
                    + pdbfile
                    + " -his -o "
                    + self.files["laststructure"]
                    + " << EOF\n"
                )
        #
        # Select the force field
        #
        fd.write("\n" + str(forcefield) + "\n\n")
        #
        # Select histidine protonation states
        if not auto_select_his:
            #
            # Histidine
            # 0: on ND1 only (4lyt)
            # 1: on NE2 only  (7lyz, 2lzm)
            # 2: Double protonated (2lzt & 4lzt)
            fd.write("1\n")
        #
        # For protonating E35 and D52 in HEWL
        #
        # fd.write('0\n0\n1\n0\n0\n0\n0\n')

        # fd.write('0\n1\n')
        #
        # Water model
        #
        fd.write("1\n\n")
        fd.write("\n\n")
        fd.close()
        status = os.system('/bin/tcsh -i -c "source ./pdb2gmx.scr"')
        if status != 0:
            raise Exception("arrg")
        return

    def center(self, dist):
        #
        # Center molecule in box and add dist space between box and protein
        #
        if not self.files.has_key("laststructure"):
            raise "No valid input file for editconf"
        status = os.system(
            "editconf -f "
            + self.files["laststructure"]
            + " -o out.gro -bt cubic -d "
            + str(dist)
        )
        self.files["laststructure"] = "out.gro"
        print "Editconf returned", status
        return

    def solvate(self):
        #
        # Solvate the system
        #
        print "Solvating system"
        status = os.system(
            "genbox -cp "
            + self.files["laststructure"]
            + " -cs -p topol.top -o solvated.gro"
        )
        self.files["laststructure"] = "solvated.gro"
        print "genbox returned:", status
        return

    def EM(self, emtol=2000, nsteps=1000, emstep=0.01, nstenergy=1):
        #
        # Define the standard em.mdp file
        #
        params = {
            "cpp": "/lib/cpp",
            "define": "-DFLEX_SPC",
            "constraints": "none",
            "integrator": "steep",
            "nsteps": str(nsteps),
            "emtol": str(emtol),
            "emstep": str(emstep),
            "nstcomm": "1",
            "ns_type": "grid",
            "rlist": "2",
            "rlist": "1",
            "rcoulomb": "1.0",
            "epsilon_r": "1000.0",
            "rvdw": "1.0",
            "Tcoupl": "no",
            "Pcoupl": "no",
            "gen_vel": "no",
            "nstenergy": str(nstenergy),
        }
        fd = open("em.mdp", "w")
        for key in params.keys():
            fd.write("%s\t\t=  %s\n" % (key, params[key]))
        fd.close()
        #
        # Preprocessing
        #
        status = os.system(
            "grompp_d -v -f em.mdp -c "
            + self.files["laststructure"]
            + " -o em.tpr -p topol.top"
        )
        print "grompp returned:", status
        #
        # Do minimisation
        #
        self.files["energytraj"] = "energy.ene"
        self.files["laststructure"] = "after_em.pdb"
        status = os.system(
            "mdrun_d -v -s em.tpr -o em.trr -c "
            + self.files["laststructure"]
            + " -g emlog -e "
            + self.files["energytraj"]
        )
        print "mdrun returned:", status
        return

    def PR_MD(self, nsteps=5000, genseed="173529", uparams={}):
        #
        # Standard pr.mdp file
        #
        params = {
            "title": "Position_restrained_MD",
            "cpp": "/lib/cpp",
            "define": "-DPOSRES",
            "constraints": "all-bonds",
            "integrator": "md",
            "tinit": "0.0",
            "dt": "0.002",
            "nsteps": "5000",
            "nstcomm": "1",
            "nstxout": "50",
            "nstvout": "1000",
            "nstfout": "0",
            "nstlog": "10",
            "nstenergy": "10",
            "nstlist": "10",
            "ns_type": "grid",
            "rlist": 1.0,
            "coulombtype": "PME",
            "fourierspacing": "0.1",
            "pme_order": "4",
            "rcoulomb": "1.0",
            "rvdw": 1.0,
            "Tcoupl": "v-rescale",
            "tc-grps": "System",
            "tau_t": "0.1",
            "ref_t": "300",
            "energygrps": "Protein SOL",
            "Pcoupl": "Berendsen",
            "tau_p": "0.5",
            "compressibility": "4.5e-5",
            "ref_p": "1.0",
            "gen_vel": "yes",
            "gen_temp": "300.0",
            "gen_seed": genseed,
        }

        params["nsteps"] = str(nsteps)
        #
        # Did the user pass anything?
        #
        for p in uparams.keys():
            params[p] = uparams[p]
            print p, params[p]
        #
        # Write the file
        fd = open("pr.mdp", "w")
        for key in params.keys():
            fd.write("%s\t\t=  %s\n" % (key, params[key]))
        fd.close()
        #
        # Preprocessing
        #
        status = os.system(
            "grompp -v -f pr.mdp -c "
            + self.files["laststructure"]
            + " -o pr.tpr -p topol.top"
        )
        print "grompp returned:", status

        #
        # Do minimisation
        #
        self.files["energytraj"] = "pr.ene"
        self.files["laststructure"] = "after_pr.pdb"
        status = os.system(
            "mdrun -v -s pr.tpr -o pr.trr -c "
            + self.files["laststructure"]
            + " -g prlog -e "
            + self.files["energytraj"]
        )
        print "mdrun returned:", status
        return

    #
    # ------
    #

    def MD(self, steps=100000, uparams={}):
        #
        # Standard full.mdp file from http://rugmd4.chem.rug.nl/~gmx/online2.0/getting_started.html#full
        #
        # Particle-Particle Particle-Mesh for long range electrostatics
        #
        params = {
            "title": "Full_MD",
            "cpp": "/lib/cpp",
            "constraints": "all-bonds",
            "integrator": "md",
            "tinit": "0.0",
            "dt": "0.002",
            "nsteps": str(steps),
            "nstcomm": "1",
            "nstxout": "250",
            "nstvout": "1000",
            "nstfout": "0",
            "nstlog": "100",
            "nstenergy": "100",
            "nstlist": "10",
            "ns_type": "grid",
            "Tcoupl": "v-rescale",
            "tc-grps": "System",
            "tau_t": "0.1",
            "ref_t": "300",
            "energygrps": "Protein SOL",
            "Pcoupl": "Berendsen",
            "tau_p": "0.5",
            "compressibility": "4.5e-5",
            "ref_p": "1.0",
            "gen_vel": "no",
            "gen_temp": "300.0",
            "gen_seed": "173529",
            # Neighbour searching
            "pbc": "xyz",
            # VDW
            "rlist": "1.0",
            "rvdw": "1.0",
            # Electrostatics
            "coulombtype": "PME",
            "fourierspacing": "0.1",
            "pme_order": "4",
            "rcoulomb": "1.0",
        }
        #
        # Did the user pass anything?
        #
        for p in uparams.keys():
            params[p] = uparams[p]
        #
        # Write the control file
        #
        fd = open("full.mdp", "w")
        for key in params.keys():
            fd.write("%s\t\t=  %s\n" % (key, params[key]))
        fd.close()
        #
        # Preprocessing
        #
        self.files["mdpfile"] = "full.mdp"
        self.files["runinput"] = "full.tpr"
        command = (
            "grompp -v -f "
            + self.files["mdpfile"]
            + " -c "
            + self.files["laststructure"]
            + " -o "
            + self.files["runinput"]
            + " -p topol.top"
        )
        print "Using command"
        print command
        status = os.system(command)
        print "grompp returned:", status
        if status != 0:
            raise Exception("Exiting with error after grompp")
        #
        # Do Molecular Dynamics
        #
        backup = self.files.copy()
        self.files["energytraj"] = "MD.ene"
        self.files["laststructure"] = "after_MD.pdb"
        self.files["trajectory"] = "MD.trr"
        self.MDtime = steps / 500.0
        command = (
            "mdrun -v -s "
            + self.files["runinput"]
            + " -o "
            + self.files["trajectory"]
            + " -c "
            + self.files["laststructure"]
            + " -g mdlog -e "
            + self.files["energytraj"]
        )
        status = os.system(command)
        print "mdrun returned:", status
        if status != 0:
            self.files = backup.copy()
        return

    #
    # ---
    #

    def get_snapshots(self, numsnapshots):
        """Get the Gromacs snapshots"""
        import os

        fd = open("trjconv.scr", "w")
        self.files["snapshots"] = "MD.snapshot.pdb"
        command = (
            "trjconv -s %s -f %s -o %s.pdb -fit rot+trans -sep -dt %3d << EOF\n"
            % (
                self.files["runinput"],
                self.files["trajectory"],
                self.files["snapshots"],
                max(1, self.MDtime / numsnapshots),
            )
        )
        fd.write(command)
        fd.write("1\n")  # Selecting protein for superpositioning
        fd.write("1\n")  # Selecting protein for output
        fd.close()
        #
        # Run the script
        #
        status = os.system('/bin/tcsh -i -c "source ./trjconv.scr"')
        if status != 0:
            raise Exception("trjconv did not run")
        #
        # Find the snapshots
        #

        files = os.listdir(os.getcwd())
        snapshots = []
        for fn in files:
            if self.files["snapshots"] in fn:
                snapshots.append(os.path.join(os.getcwd(), fn))
        if len(snapshots) < numsnapshots:
            print "Got only %d snapshots..." % (len(snapshots))
        return snapshots
